# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""TensorFlowTaskRunner module."""

import numpy as np
import tensorflow.compat.v1 as tf
from tqdm import tqdm

from openfl.utilities import split_tensor_dict_for_holdouts
from openfl.utilities import TensorKey
from .runner import TaskRunner


class TensorFlowTaskRunner(TaskRunner):
    """
    Base class for TensorFlow models in the Federated Learning solution.

        child classes should have __init__ function signature (self, data, kwargs),
        and should overwrite at least the following while defining the model
    """

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        tf.disable_v2_behavior()

        super().__init__(**kwargs)

        self.assign_ops = None
        self.placeholders = None

        self.tvar_assign_ops = None
        self.tvar_placeholders = None

        # construct the shape needed for the input features
        self.input_shape = (None,) + self.data_loader.get_feature_shape()

        # Required tensorkeys for all public functions in TensorFlowTaskRunner
        self.required_tensorkeys_for_function = {}

        # Required tensorkeys for all public functions in TensorFlowTaskRunner
        self.required_tensorkeys_for_function = {}

        # tensorflow session
        self.sess = None
        # input featrures to the model
        self.X = None
        # input labels to the model
        self.y = None
        # optimizer train step operation
        self.train_step = None
        # model loss function
        self.loss = None
        # model output tensor
        self.output = None
        # function used to validate the model outputs against labels
        self.validation_metric = None
        # tensorflow trainable variables
        self.tvars = None
        # self.optimizer.variables() once self.optimizer is defined
        self.opt_vars = None
        # self.tvars + self.opt_vars
        self.fl_vars = None

    def rebuild_model(self, round_num, input_tensor_dict, validation=False):
        """
        Parse tensor names and update weights of model. Handles the optimizer treatment.

        Returns:
            None
        """
        if self.opt_treatment == 'RESET':
            self.reset_opt_vars()
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)
        elif (round_num > 0 and self.opt_treatment == 'CONTINUE_GLOBAL'
              and not validation):
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=True)
        else:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)

    def train_batches(self, col_name, round_num, input_tensor_dict,
                      epochs=1, use_tqdm=False, **kwargs):
        """
        Perform the training.

        Is expected to perform draws randomly, without replacement until data is exausted. Then
        data is replaced and shuffled and draws continue.

        Args:
            use_tqdm (bool): True = use tqdm to print a progress
             bar (Default=False)
            epochs (int): Number of epochs to train
        Returns:
            float: loss metric
        """
        batch_size = self.data_loader.batch_size

        if kwargs['batch_size']:
            batch_size = kwargs['batch_size']

        # rebuild model with updated weights
        self.rebuild_model(round_num, input_tensor_dict)

        tf.keras.backend.set_learning_phase(True)
        losses = []

        for epoch in range(epochs):
            self.logger.info(f'Run {epoch} epoch of {round_num} round')
            # get iterator for batch draws (shuffling happens here)
            gen = self.data_loader.get_train_loader(batch_size)
            if use_tqdm:
                gen = tqdm.tqdm(gen, desc='training epoch')

            for (X, y) in gen:
                losses.append(self.train_batch(X, y))

        # Output metric tensors (scalar)
        origin = col_name
        tags = ('trained',)
        output_metric_dict = {
            TensorKey(
                self.loss_name, origin, round_num, True, ('metric',)
            ): np.array(np.mean(losses))
        }

        # output model tensors (Doesn't include TensorKey)
        output_model_dict = self.get_tensor_dict(with_opt_vars=True)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            self.logger, output_model_dict,
            **self.tensor_dict_split_fn_kwargs
        )

        # Create global tensorkeys
        global_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags):
                nparray for tensor_name, nparray in global_model_dict.items()
        }
        # Create tensorkeys that should stay local
        local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags):
                nparray for tensor_name, nparray in local_model_dict.items()
        }
        # The train/validate aggregated function of the next round will
        # look for the updated model parameters.
        # This ensures they will be resolved locally
        next_local_tensorkey_model_dict = {
            TensorKey(
                tensor_name, origin, round_num + 1, False, ('model',)
            ): nparray for tensor_name, nparray in local_model_dict.items()}

        global_tensor_dict = {
            **output_metric_dict,
            **global_tensorkey_model_dict
        }
        local_tensor_dict = {
            **local_tensorkey_model_dict,
            **next_local_tensorkey_model_dict
        }

        # Update the required tensors if they need to be pulled from
        # the aggregator
        # TODO this logic can break if different collaborators have different
        #  roles between rounds.
        # For example, if a collaborator only performs validation in the first
        # round but training in the second, it has no way of knowing the
        # optimizer state tensor names to request from the aggregator because
        # these are only created after training occurs. A work around could
        # involve doing a single epoch of training on random data to get the
        # optimizer names, and then throwing away the model.
        if self.opt_treatment == 'CONTINUE_GLOBAL':
            self.initialize_tensorkeys_for_functions(with_opt_vars=True)

        return global_tensor_dict, local_tensor_dict

    def train_batch(self, X, y):
        """
        Train the model on a single batch.

        Args:
            X: Input to the model
            y: Ground truth label to the model

        Returns:
            float: loss metric
        """
        feed_dict = {self.X: X, self.y: y}

        # run the train step and return the loss
        _, loss = self.sess.run([self.train_step, self.loss], feed_dict=feed_dict)

        return loss

    def validate(self, col_name, round_num,
                 input_tensor_dict, use_tqdm=False, **kwargs):
        """
        Run validation.

        Returns:
            dict: {<metric>: <value>}
        """
        batch_size = self.data_loader.batch_size

        if kwargs['batch_size']:
            batch_size = kwargs['batch_size']

        self.rebuild_model(round_num, input_tensor_dict, validation=True)

        tf.keras.backend.set_learning_phase(False)

        score = 0

        gen = self.data_loader.get_valid_loader(batch_size)
        if use_tqdm:
            gen = tqdm.tqdm(gen, desc='validating')

        for X, y in gen:
            weight = X.shape[0] / self.data_loader.get_valid_data_size()
            _, s = self.validate_batch(X, y)
            score += s * weight

        origin = col_name
        suffix = 'validate'
        if kwargs['apply'] == 'local':
            suffix += '_local'
        else:
            suffix += '_agg'
        tags = ('metric', suffix)
        output_tensor_dict = {
            TensorKey(
                self.validation_metric_name, origin, round_num, True, tags
            ): np.array(score)}

        # return empty dict for local metrics
        return output_tensor_dict, {}

    def validate_batch(self, X, y):
        """Validate the model on a single local batch.

        Args:
            X: Input to the model
            y: Ground truth label to the model

        Returns:
            float: loss metric

        """
        feed_dict = {self.X: X, self.y: y}

        return self.sess.run(
            [self.output, self.validation_metric], feed_dict=feed_dict)

    def get_tensor_dict(self, with_opt_vars=True):
        """Get the dictionary weights.

        Get the weights from the tensor

        Args:
            with_opt_vars (bool): Specify if we also want to get the variables
             of the optimizer

        Returns:
            dict: The weight dictionary {<tensor_name>: <value>}

        """
        if with_opt_vars is True:
            variables = self.fl_vars
        else:
            variables = self.tvars

        # FIXME: do this in one call?
        return {var.name: val for var, val in zip(
            variables, self.sess.run(variables))}

    def set_tensor_dict(self, tensor_dict, with_opt_vars):
        """Set the tensor dictionary.

        Set the model weights with a tensor
         dictionary: {<tensor_name>: <value>}.

        Args:
            tensor_dict (dict): The model weights dictionary
            with_opt_vars (bool): Specify if we also want to set the variables
             of the optimizer

        Returns:
            None
        """
        if with_opt_vars:
            self.assign_ops, self.placeholders = tf_set_tensor_dict(
                tensor_dict, self.sess, self.fl_vars,
                self.assign_ops, self.placeholders
            )
        else:
            self.tvar_assign_ops, self.tvar_placeholders = tf_set_tensor_dict(
                tensor_dict,
                self.sess,
                self.tvars,
                self.tvar_assign_ops,
                self.tvar_placeholders
            )

    def reset_opt_vars(self):
        """Reinitialize the optimizer variables."""
        for v in self.opt_vars:
            v.initializer.run(session=self.sess)

    def initialize_globals(self):
        """Initialize Global Variables.

        Initialize all global variables

        Returns:
            None
        """
        self.sess.run(tf.global_variables_initializer())

    def _get_weights_names(self, with_opt_vars=True):
        """Get the weights.

        Args:
            with_opt_vars (bool): Specify if we also want to get the variables
             of the optimizer.

        Returns:
            list : The weight names list
        """
        if with_opt_vars is True:
            variables = self.fl_vars
        else:
            variables = self.tvars

        return [var.name for var in variables]

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """
        Get the required tensors for specified function that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.

        Returns:
            list : [TensorKey]
        """
        if func_name == 'validate':
            local_model = 'apply=' + str(kwargs['apply'])
            return self.required_tensorkeys_for_function[func_name][local_model]
        else:
            return self.required_tensorkeys_for_function[func_name]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """
        Set the required tensors for all publicly accessible methods \
            that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function

        """
        # TODO there should be a way to programmatically iterate through
        #  all of the methods in the class and declare the tensors.
        # For now this is done manually

        output_model_dict = self.get_tensor_dict(with_opt_vars=with_opt_vars)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            self.logger, output_model_dict,
            **self.tensor_dict_split_fn_kwargs
        )
        if not with_opt_vars:
            global_model_dict_val = global_model_dict
            local_model_dict_val = local_model_dict
        else:
            output_model_dict = self.get_tensor_dict(with_opt_vars=False)
            global_model_dict_val, local_model_dict_val = split_tensor_dict_for_holdouts(
                self.logger,
                output_model_dict,
                **self.tensor_dict_split_fn_kwargs
            )

        self.required_tensorkeys_for_function['train_batches'] = [
            TensorKey(tensor_name, 'GLOBAL', 0, False, ('model',))
            for tensor_name in global_model_dict]
        self.required_tensorkeys_for_function['train_batches'] += [
            TensorKey(tensor_name, 'LOCAL', 0, False, ('model',))
            for tensor_name in local_model_dict]

        # Validation may be performed on local or aggregated (global)
        # model, so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function['validate'] = {}
        # TODO This is not stateless. The optimizer will not be
        self.required_tensorkeys_for_function['validate']['apply=local'] = [
            TensorKey(tensor_name, 'LOCAL', 0, False, ('trained',))
            for tensor_name in {
                **global_model_dict_val,
                **local_model_dict_val
            }
        ]
        self.required_tensorkeys_for_function['validate']['apply=global'] = [
            TensorKey(tensor_name, 'GLOBAL', 0, False, ('model',))
            for tensor_name in global_model_dict_val
        ]
        self.required_tensorkeys_for_function['validate']['apply=global'] += [
            TensorKey(tensor_name, 'LOCAL', 0, False, ('model',))
            for tensor_name in local_model_dict_val
        ]


# FIXME: what's a nicer construct than this? ugly interface. Perhaps we
#  get an object with an assumed interface that lets is set/get these?
# Note that this will return the assign_ops and placeholder nodes it uses
# if called with None, it will create them.
# to avoid inflating the graph, caller should keep these and pass them back
# What if we want to set a different group of vars in the middle?
# It is good if it is the subset of the original variables.
def tf_set_tensor_dict(tensor_dict, session, variables,
                       assign_ops=None, placeholders=None):
    """Tensorflow set tensor dictionary.

    Args:
        tensor_dict: Dictionary of tensors
        session: TensorFlow session
        variables: TensorFlow variables
        assign_ops: TensorFlow operations (Default=None)
        placeholders: TensorFlow placeholders (Default=None)

    Returns:
        assign_ops, placeholders

    """
    if placeholders is None:
        placeholders = {
            v.name: tf.placeholder(v.dtype, shape=v.shape) for v in variables
        }
    if assign_ops is None:
        assign_ops = {
            v.name: tf.assign(v, placeholders[v.name]) for v in variables
        }

    for k, v in tensor_dict.items():
        session.run(assign_ops[k], feed_dict={placeholders[k]: v})

    return assign_ops, placeholders



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""AggregatorGRPCServer module."""

import logging
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from time import sleep

from grpc import server
from grpc import ssl_server_credentials

from openfl.protocols import aggregator_pb2
from openfl.protocols import aggregator_pb2_grpc
from openfl.protocols import utils
from openfl.utilities import check_equal
from openfl.utilities import check_is_in

from .grpc_channel_options import channel_options

logger = logging.getLogger(__name__)


class AggregatorGRPCServer(aggregator_pb2_grpc.AggregatorServicer):
    """gRPC server class for the Aggregator."""

    def __init__(self,
                 aggregator,
                 agg_port,
                 tls=True,
                 disable_client_auth=False,
                 root_certificate=None,
                 certificate=None,
                 private_key=None,
                 **kwargs):
        """
        Class initializer.

        Args:
            aggregator: The aggregator
        Args:
            fltask (FLtask): The gRPC service task.
            tls (bool): To disable the TLS. (Default: True)
            disable_client_auth (bool): To disable the client side
            authentication. (Default: False)
            root_certificate (str): File path to the CA certificate.
            certificate (str): File path to the server certificate.
            private_key (str): File path to the private key.
            kwargs (dict): Additional arguments to pass into function
        """
        self.aggregator = aggregator
        self.uri = f'[::]:{agg_port}'
        self.tls = tls
        self.disable_client_auth = disable_client_auth
        self.root_certificate = root_certificate
        self.certificate = certificate
        self.private_key = private_key
        self.server = None
        self.server_credentials = None

        self.logger = logging.getLogger(__name__)

    def validate_collaborator(self, request, context):
        """
        Validate the collaborator.

        Args:
            request: The gRPC message request
            context: The gRPC context

        Raises:
            ValueError: If the collaborator or collaborator certificate is not
             valid then raises error.

        """
        if self.tls:
            common_name = context.auth_context()[
                'x509_common_name'][0].decode('utf-8')
            collaborator_common_name = request.header.sender
            if not self.aggregator.valid_collaborator_cn_and_id(
                    common_name, collaborator_common_name):
                raise ValueError(
                    f'Invalid collaborator. CN: |{common_name}| '
                    f'collaborator_common_name: |{collaborator_common_name}|')

    def get_header(self, collaborator_name):
        """
        Compose and return MessageHeader.

        Args:
            collaborator_name : str
                The collaborator the message is intended for
        """
        return aggregator_pb2.MessageHeader(
            sender=self.aggregator.uuid,
            receiver=collaborator_name,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name
        )

    def check_request(self, request):
        """
        Validate request header matches expected values.

        Args:
            request : protobuf
                Request sent from a collaborator that requires validation
        """
        # TODO improve this check. the sender name could be spoofed
        check_is_in(request.header.sender, self.aggregator.authorized_cols, self.logger)

        # check that the message is for me
        check_equal(request.header.receiver, self.aggregator.uuid, self.logger)

        # check that the message is for my federation
        check_equal(
            request.header.federation_uuid, self.aggregator.federation_uuid, self.logger)

        # check that we agree on the single cert common name
        check_equal(
            request.header.single_col_cert_common_name,
            self.aggregator.single_col_cert_common_name,
            self.logger
        )

    def GetTasks(self, request, context):  # NOQA:N802
        """
        Request a job from aggregator.

        Args:
            request: The gRPC message request
            context: The gRPC context

        """
        self.validate_collaborator(request, context)
        self.check_request(request)
        collaborator_name = request.header.sender
        tasks, round_number, sleep_time, time_to_quit = self.aggregator.get_tasks(
            request.header.sender)
        if tasks:
            if isinstance(tasks[0], str):
                # backward compatibility
                tasks_proto = [
                    aggregator_pb2.Task(
                        name=task,
                    ) for task in tasks
                ]
            else:
                tasks_proto = [
                    aggregator_pb2.Task(
                        name=task.name,
                        function_name=task.function_name,
                        task_type=task.task_type,
                        apply_local=task.apply_local
                    ) for task in tasks
                ]
        else:
            tasks_proto = []

        return aggregator_pb2.GetTasksResponse(
            header=self.get_header(collaborator_name),
            round_number=round_number,
            tasks=tasks_proto,
            sleep_time=sleep_time,
            quit=time_to_quit
        )

    def GetTensor(self, request, context):  # NOQA:N802
        """
        Request a job from aggregator.

        Args:
            request: The gRPC message request
            context: The gRPC context

        """
        self.validate_collaborator(request, context)
        self.check_request(request)
        collaborator_name = request.header.sender
        tensor_name = request.tensor_name
        require_lossless = request.require_lossless
        round_number = request.round_number
        report = request.report
        aggregated = request.aggregated
        tags = tuple(request.tags)

        named_tensor = self.aggregator.get_tensor(
            collaborator_name, tensor_name, round_number, report, tags, require_lossless, aggregated)

        return aggregator_pb2.GetTensorResponse(
            header=self.get_header(collaborator_name),
            round_number=round_number,
            tensor=named_tensor
        )

    def GetSynch(self, request, context):  # NOQA:N802
        """
        Request synchronization for a task.
        Args:
            request: The gRPC message request
            context: The gRPC context
        """
        self.validate_collaborator(request, context)
        self.check_request(request)
        collaborator_name = request.header.sender
        task_name = request.task_name
        round_number = request.round_number
        is_completed = self.aggregator.synch(task_name, round_number)

        return aggregator_pb2.SynchResponse(
            header=self.get_header(collaborator_name),
            is_completed=is_completed
        )

    def SendLocalTaskResults(self, request, context):  # NOQA:N802
        """
        Request a model download from aggregator.

        Args:
            request: The gRPC message request
            context: The gRPC context

        """
        try:
            proto = aggregator_pb2.TaskResults()
            proto = utils.datastream_to_proto(proto, request)
        except RuntimeError:
            raise RuntimeError(
                'Empty stream message, reestablishing connection from client to resume training...'
            )

        self.validate_collaborator(proto, context)
        # all messages get sanity checked
        self.check_request(proto)

        collaborator_name = proto.header.sender
        task_name = proto.task_name
        round_number = proto.round_number
        data_size = proto.data_size
        named_tensors = proto.tensors
        self.aggregator.send_local_task_results(
            collaborator_name, round_number, task_name, data_size, named_tensors)
        # turn data stream into local model update
        return aggregator_pb2.SendLocalTaskResultsResponse(
            header=self.get_header(collaborator_name)
        )

    def get_server(self):
        """Return gRPC server."""
        self.server = server(ThreadPoolExecutor(max_workers=cpu_count()),
                             options=channel_options)

        aggregator_pb2_grpc.add_AggregatorServicer_to_server(self, self.server)

        if not self.tls:

            self.logger.warn(
                'gRPC is running on insecure channel with TLS disabled.')
            port = self.server.add_insecure_port(self.uri)
            self.logger.info(f'Insecure port: {port}')

        else:

            with open(self.private_key, 'rb') as f:
                private_key_b = f.read()
            with open(self.certificate, 'rb') as f:
                certificate_b = f.read()
            with open(self.root_certificate, 'rb') as f:
                root_certificate_b = f.read()

            if self.disable_client_auth:
                self.logger.warn('Client-side authentication is disabled.')

            self.server_credentials = ssl_server_credentials(
                ((private_key_b, certificate_b),),
                root_certificates=root_certificate_b,
                require_client_auth=not self.disable_client_auth
            )

            self.server.add_secure_port(self.uri, self.server_credentials)

        return self.server

    def serve(self):
        """Start an aggregator gRPC service."""
        self.get_server()

        self.logger.info('Starting Aggregator gRPC Server')
        self.server.start()

        try:
            while not self.aggregator.all_quit_jobs_sent():
                sleep(5)
        except KeyboardInterrupt:
            pass

        self.server.stop(0)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Keras Framework Adapter plugin."""
from logging import getLogger

from .framework_adapter_interface import FrameworkAdapterPluginInterface

logger = getLogger(__name__)
class FrameworkAdapterPlugin(FrameworkAdapterPluginInterface):
    """Framework adapter plugin class."""

    def __init__(self) -> None:
        """Initialize framework adapter."""
        pass

    @staticmethod
    def serialization_setup():
        """Prepare model for serialization (optional)."""
        # Source: https://github.com/tensorflow/tensorflow/issues/34697
        import tensorflow as tf
        from tensorflow.keras.models import Model
        from tensorflow.python.keras.layers import deserialize
        from tensorflow.python.keras.layers import serialize
        from tensorflow.python.keras.saving import saving_utils

        def unpack(model, training_config, weights):
            restored_model = deserialize(model)
            if training_config is not None:
                restored_model.compile(
                    **saving_utils.compile_args_from_training_config(
                        training_config
                    )
                )
            restored_model.set_weights(weights)
            return restored_model

        # Hotfix function, not required for TF versions above 2.7.1.
        # https://github.com/keras-team/keras/pull/14748.
        def make_keras_picklable():

            def __reduce__(self):  # NOQA:N807
                model_metadata = saving_utils.model_metadata(self)
                training_config = model_metadata.get('training_config', None)
                model = serialize(self)
                weights = self.get_weights()
                return (unpack, (model, training_config, weights))

            cls = Model
            cls.__reduce__ = __reduce__

        # Run the function
        if tf.__version__ <= '2.7.1':
            logger.warn(f'Applying hotfix for model serialization.'
            'Please consider updating to tensorflow>=2.8 to silence this warning.')
            make_keras_picklable()

    @staticmethod
    def get_tensor_dict(model, optimizer=None, suffix=''):
        """
        Extract tensor dict from a model and an optimizer.

        Returns:
        dict {weight name: numpy ndarray}
        """
        model_weights = _get_weights_dict(model, suffix)

        if optimizer is not None:
            opt_weights = _get_weights_dict(optimizer, suffix)

            model_weights.update(opt_weights)
            if len(opt_weights) == 0:
                # ToDo: warn user somehow
                pass

        return model_weights

    @staticmethod
    def set_tensor_dict(model, tensor_dict, optimizer=None, device='cpu'):
        """
        Set the model weights with a tensor dictionary.

        Args:
            tensor_dict: the tensor dictionary
            with_opt_vars (bool): True = include the optimizer's status.
        """
        model_weight_names = [weight.name for weight in model.weights]
        model_weights_dict = {
            name: tensor_dict[name] for name in model_weight_names
        }
        _set_weights_dict(model, model_weights_dict)

        if optimizer is not None:
            opt_weight_names = [
                weight.name for weight in optimizer.weights
            ]
            opt_weights_dict = {
                name: tensor_dict[name] for name in opt_weight_names
            }
            _set_weights_dict(optimizer, opt_weights_dict)


def _get_weights_dict(obj, suffix=''):
    """
    Get the dictionary of weights.

    Parameters
    ----------
    obj : Model or Optimizer
        The target object that we want to get the weights.

    Returns
    -------
    dict
        The weight dictionary.
    """
    weights_dict = {}
    weight_names = [weight.name for weight in obj.weights]
    weight_values = obj.get_weights()
    for name, value in zip(weight_names, weight_values):
        weights_dict[name + suffix] = value
    return weights_dict


def _set_weights_dict(obj, weights_dict):
    """Set the object weights with a dictionary.

    The obj can be a model or an optimizer.

    Args:
        obj (Model or Optimizer): The target object that we want to set
        the weights.
        weights_dict (dict): The weight dictionary.

    Returns:
        None
    """
    weight_names = [weight.name for weight in obj.weights]
    weight_values = [weights_dict[name] for name in weight_names]
    obj.set_weights(weight_values)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Adam optimizer module."""

from typing import Dict
from typing import Optional
from typing import Tuple

import numpy as np

from .adam_optimizer import NumPyAdam


class NumPyYogi(NumPyAdam):
    """Yogi optimizer implementation.

    Original paper:
    https://papers.nips.cc/paper/8186-adaptive-methods-for-nonconvex-optimization
    """

    def __init__(
        self,
        *,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        betas: Tuple[float, float] = (0.9, 0.999),
        initial_accumulator_value: float = 0.0,
        epsilon: float = 1e-8,
    ) -> None:
        """Initialize.

        Args:
            params: Parameters to be stored for optimization.
            model_interface: Model interface instance to provide parameters.
            learning_rate: Tuning parameter that determines
                the step size at each iteration.
            betas: Coefficients used for computing running
                averages of gradient and its square.
            initial_accumulator_value: Initial value for gradients
                and squared gradients.
            epsilon: Value for computational stability.
        """
        super().__init__(params=params,
                         model_interface=model_interface,
                         learning_rate=learning_rate,
                         betas=betas,
                         initial_accumulator_value=initial_accumulator_value,
                         epsilon=epsilon)

    def _update_second_moment(self, grad_name: str, grad: np.ndarray) -> None:
        """Override second moment update rule for Yogi optimization updates."""
        sign = np.sign(grad**2 - self.grads_second_moment[grad_name])
        self.grads_second_moment[grad_name] = (self.beta_2
                                               * self.grads_second_moment[grad_name]
                                               + (1.0 - self.beta_2) * sign * grad**2)

    def step(self, gradients: Dict[str, np.ndarray]) -> None:
        """
        Perform a single step for parameter update.

        Implement Yogi optimizer weights update rule.

        Args:
            gradients: Partial derivatives with respect to optimized parameters.
        """
        super().step(gradients)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Task package."""

import pkgutil
from warnings import catch_warnings
from warnings import simplefilter

with catch_warnings():
    simplefilter(action='ignore', category=FutureWarning)
    if pkgutil.find_loader('tensorflow'):
        # ignore deprecation warnings in command-line interface
        import tensorflow  # NOQA

from .runner import TaskRunner  # NOQA


if pkgutil.find_loader('tensorflow'):
    from .runner_tf import TensorFlowTaskRunner  # NOQA
    from .runner_keras import KerasTaskRunner  # NOQA
    from .fl_model import FederatedModel  # NOQA
if pkgutil.find_loader('torch'):
    from .runner_pt import PyTorchTaskRunner  # NOQA
    from .fl_model import FederatedModel  # NOQA
if pkgutil.find_loader('torch') and pkgutil.find_loader('tensorflow'):
    from .runner_fe import FastEstimatorTaskRunner  # NOQA



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Mnist Shard Descriptor."""

import logging
import os
from typing import List

import numpy as np
import requests

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)


class MnistShardDataset(ShardDataset):
    """Mnist Shard dataset class."""

    def __init__(self, x, y, data_type, rank=1, worldsize=1):
        """Pick rank-specific subset of (x, y)"""
        self.data_type = data_type
        self.rank = rank
        self.worldsize = worldsize
        self.x = x[self.rank - 1::self.worldsize]
        self.y = y[self.rank - 1::self.worldsize]

    def __getitem__(self, index: int):
        """Return an item by the index."""
        return self.x[index], self.y[index]

    def __len__(self):
        """Return the len of the dataset."""
        return len(self.x)


class MnistShardDescriptor(ShardDescriptor):
    """Mnist Shard descriptor class."""

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            **kwargs
    ):
        """Initialize MnistShardDescriptor."""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        (x_train, y_train), (x_test, y_test) = self.download_data()
        self.data_by_type = {
            'train': (x_train, y_train),
            'val': (x_test, y_test)
        }

    def get_shard_dataset_types(self) -> List[str]:
        """Get available shard dataset types."""
        return list(self.data_by_type)

    def get_dataset(self, dataset_type='train'):
        """Return a shard dataset by type."""
        if dataset_type not in self.data_by_type:
            raise Exception(f'Wrong dataset type: {dataset_type}')
        return MnistShardDataset(
            *self.data_by_type[dataset_type],
            data_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return ['28', '28', '1']

    @property
    def target_shape(self):
        """Return the target shape info."""
        return ['1']

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Mnist dataset, shard number {self.rank}'
                f' out of {self.worldsize}')

    def download_data(self):
        """Download prepared dataset."""
        local_file_path = 'mnist.npz'
        mnist_url = 'https://storage.googleapis.com/tensorflow/tf-keras-datasets/mnist.npz'
        response = requests.get(mnist_url)
        with open(local_file_path, 'wb') as f:
            f.write(response.content)

        with np.load(local_file_path) as f:
            x_train, y_train = f['x_train'], f['y_train']
            x_test, y_test = f['x_test'], f['y_test']
            x_train = np.reshape(x_train, (-1, 28, 28, 1))
            x_test = np.reshape(x_test, (-1, 28, 28, 1))

        os.remove(local_file_path)  # remove mnist.npz
        print('Mnist data was loaded!')
        return (x_train, y_train), (x_test, y_test)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""PyTorch FedProx optimizer module."""

import math

import torch
from torch.optim import Optimizer
from torch.optim.optimizer import required


class FedProxOptimizer(Optimizer):
    """FedProx optimizer.

    Paper: https://arxiv.org/pdf/1812.06127.pdf
    """

    def __init__(self,
                 params,
                 lr=required,
                 mu=0.0,
                 momentum=0,
                 dampening=0,
                 weight_decay=0,
                 nesterov=False):
        """Initialize."""
        if momentum < 0.0:
            raise ValueError(f'Invalid momentum value: {momentum}')
        if lr is not required and lr < 0.0:
            raise ValueError(f'Invalid learning rate: {lr}')
        if weight_decay < 0.0:
            raise ValueError(f'Invalid weight_decay value: {weight_decay}')
        if mu < 0.0:
            raise ValueError(f'Invalid mu value: {mu}')
        defaults = {
            'dampening': dampening,
            'lr': lr,
            'momentum': momentum,
            'mu': mu,
            'nesterov': nesterov,
            'weight_decay': weight_decay,
        }

        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError('Nesterov momentum requires a momentum and zero dampening')

        super(FedProxOptimizer, self).__init__(params, defaults)

    def __setstate__(self, state):
        """Set optimizer state."""
        super(FedProxOptimizer, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    @torch.no_grad()
    def step(self, closure=None):
        """Perform a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']
            mu = group['mu']
            w_old = group['w_old']
            for p, w_old_p in zip(group['params'], w_old):
                if p.grad is None:
                    continue
                d_p = p.grad
                if weight_decay != 0:
                    d_p = d_p.add(p, alpha=weight_decay)
                if momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf
                if w_old is not None:
                    d_p.add_(p - w_old_p, alpha=mu)
                p.add_(d_p, alpha=-group['lr'])

        return loss

    def set_old_weights(self, old_weights):
        """Set the global weights parameter to `old_weights` value."""
        for param_group in self.param_groups:
            param_group['w_old'] = old_weights


class FedProxAdam(Optimizer):
    """FedProxAdam optimizer."""

    def __init__(self, params, mu=0, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, amsgrad=False):
        """Initialize."""
        if not 0.0 <= lr:
            raise ValueError(f'Invalid learning rate: {lr}')
        if not 0.0 <= eps:
            raise ValueError(f'Invalid epsilon value: {eps}')
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f'Invalid beta parameter at index 0: {betas[0]}')
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f'Invalid beta parameter at index 1: {betas[1]}')
        if not 0.0 <= weight_decay:
            raise ValueError(f'Invalid weight_decay value: {weight_decay}')
        if mu < 0.0:
            raise ValueError(f'Invalid mu value: {mu}')
        defaults = {'lr': lr, 'betas': betas, 'eps': eps,
                    'weight_decay': weight_decay, 'amsgrad': amsgrad, 'mu': mu}
        super(FedProxAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        """Set optimizer state."""
        super(FedProxAdam, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    def set_old_weights(self, old_weights):
        """Set the global weights parameter to `old_weights` value."""
        for param_group in self.param_groups:
            param_group['w_old'] = old_weights

    @torch.no_grad()
    def step(self, closure=None):
        """Perform a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            exp_avg_sqs = []
            max_exp_avg_sqs = []
            state_steps = []

            for p in group['params']:
                if p.grad is not None:
                    params_with_grad.append(p)
                    if p.grad.is_sparse:
                        raise RuntimeError(
                            'Adam does not support sparse gradients, '
                            'please consider SparseAdam instead')
                    grads.append(p.grad)

                    state = self.state[p]
                    # Lazy state initialization
                    if len(state) == 0:
                        state['step'] = 0
                        # Exponential moving average of gradient values
                        state['exp_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                        # Exponential moving average of squared gradient values
                        state['exp_avg_sq'] = torch.zeros_like(
                            p, memory_format=torch.preserve_format)
                        if group['amsgrad']:
                            # Maintains max of all exp. moving avg. of sq. grad. values
                            state['max_exp_avg_sq'] = torch.zeros_like(
                                p, memory_format=torch.preserve_format)

                    exp_avgs.append(state['exp_avg'])
                    exp_avg_sqs.append(state['exp_avg_sq'])

                    if group['amsgrad']:
                        max_exp_avg_sqs.append(state['max_exp_avg_sq'])

                    # update the steps for each param group update
                    state['step'] += 1
                    # record the step after step update
                    state_steps.append(state['step'])

            beta1, beta2 = group['betas']
            self.adam(params_with_grad,
                      grads,
                      exp_avgs,
                      exp_avg_sqs,
                      max_exp_avg_sqs,
                      state_steps,
                      group['amsgrad'],
                      beta1,
                      beta2,
                      group['lr'],
                      group['weight_decay'],
                      group['eps'],
                      group['mu'],
                      group['w_old']
                      )
        return loss

    def adam(self, params,
             grads,
             exp_avgs,
             exp_avg_sqs,
             max_exp_avg_sqs,
             state_steps,
             amsgrad,
             beta1: float,
             beta2: float,
             lr: float,
             weight_decay: float,
             eps: float,
             mu: float,
             w_old):
        """Updtae optimizer parameters."""
        for i, param in enumerate(params):
            w_old_p = w_old[i]
            grad = grads[i]
            grad.add_(param - w_old_p, alpha=mu)
            exp_avg = exp_avgs[i]
            exp_avg_sq = exp_avg_sqs[i]
            step = state_steps[i]

            bias_correction1 = 1 - beta1 ** step
            bias_correction2 = 1 - beta2 ** step

            if weight_decay != 0:
                grad = grad.add(param, alpha=weight_decay)

            # Decay the first and second moment running average coefficient
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
            if amsgrad:
                # Maintains the maximum of all 2nd moment running avg. till now
                torch.maximum(max_exp_avg_sqs[i], exp_avg_sq, out=max_exp_avg_sqs[i])
                # Use the max. for normalizing running avg. of gradient
                denom = (max_exp_avg_sqs[i].sqrt() / math.sqrt(bias_correction2)).add_(eps)
            else:
                denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(eps)

            step_size = lr / bias_correction1

            param.addcdiv_(exp_avg, denom, value=-step_size)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Click types module."""

import click

from openfl.utilities import utils


class FqdnParamType(click.ParamType):
    """Domain Type for click arguments."""

    name = 'fqdn'

    def convert(self, value, param, ctx):
        """Validate value, if value is valid, return it."""
        if not utils.is_fqdn(value):
            self.fail(f'{value} is not a valid domain name', param, ctx)
        return value


class IpAddressParamType(click.ParamType):
    """IpAddress Type for click arguments."""

    name = 'IpAddress type'

    def convert(self, value, param, ctx):
        """Validate value, if value is valid, return it."""
        if not utils.is_api_adress(value):
            self.fail(f'{value} is not a valid ip adress name', param, ctx)
        return value


FQDN = FqdnParamType()
IP_ADDRESS = IpAddressParamType()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""FederatedFastEstimator module."""

import os
from logging import getLogger
from pathlib import Path

import openfl.native as fx
from openfl.federated import Plan
from openfl.federated.data import FastEstimatorDataLoader
from openfl.federated.task import FastEstimatorTaskRunner
from openfl.protocols import utils
from openfl.utilities import split_tensor_dict_for_holdouts


class FederatedFastEstimator:
    """A wrapper for fastestimator.estimator that allows running in federated mode."""

    def __init__(self, estimator, override_config: dict = None, **kwargs):
        """Initialize."""
        self.estimator = estimator
        self.logger = getLogger(__name__)
        fx.init(**kwargs)
        if override_config:
            fx.update_plan(override_config)

    def fit(self):
        """Run the estimator."""
        import fastestimator as fe
        from fastestimator.trace.io.best_model_saver import BestModelSaver
        from sys import path

        file = Path(__file__).resolve()
        # interface root, containing command modules
        root = file.parent.resolve()
        work = Path.cwd().resolve()

        path.append(str(root))
        path.insert(0, str(work))

        # TODO: Fix this implementation. The full plan parsing is reused here,
        # but the model and data will be overwritten based on
        # user specifications
        plan_config = (Path(fx.WORKSPACE_PREFIX) / 'plan' / 'plan.yaml')
        cols_config = (Path(fx.WORKSPACE_PREFIX) / 'plan' / 'cols.yaml')
        data_config = (Path(fx.WORKSPACE_PREFIX) / 'plan' / 'data.yaml')

        plan = Plan.parse(plan_config_path=plan_config,
                          cols_config_path=cols_config,
                          data_config_path=data_config)

        self.rounds = plan.config['aggregator']['settings']['rounds_to_train']
        data_loader = FastEstimatorDataLoader(self.estimator.pipeline)
        runner = FastEstimatorTaskRunner(
            self.estimator, data_loader=data_loader)
        # Overwrite plan values
        tensor_pipe = plan.get_tensor_pipe()
        # Initialize model weights
        init_state_path = plan.config['aggregator']['settings'][
            'init_state_path']
        tensor_dict, holdout_params = split_tensor_dict_for_holdouts(
            self.logger, runner.get_tensor_dict(False))

        model_snap = utils.construct_model_proto(tensor_dict=tensor_dict,
                                                 round_number=0,
                                                 tensor_pipe=tensor_pipe)

        self.logger.info(f'Creating Initial Weights File'
                         f'    🠆 {init_state_path}')

        utils.dump_proto(model_proto=model_snap, fpath=init_state_path)

        self.logger.info('Starting Experiment...')

        aggregator = plan.get_aggregator()

        model_states = {
            collaborator: None for collaborator in plan.authorized_cols
        }
        runners = {}
        save_dir = {}
        data_path = 1
        for col in plan.authorized_cols:
            data = self.estimator.pipeline.data
            train_data, eval_data, test_data = split_data(
                data['train'], data['eval'], data['test'],
                data_path, len(plan.authorized_cols))
            pipeline_kwargs = {}
            for k, v in self.estimator.pipeline.__dict__.items():
                if k in ['batch_size', 'ops', 'num_process',
                         'drop_last', 'pad_value', 'collate_fn']:
                    pipeline_kwargs[k] = v
            pipeline_kwargs.update({
                'train_data': train_data,
                'eval_data': eval_data,
                'test_data': test_data
            })
            pipeline = fe.Pipeline(**pipeline_kwargs)

            data_loader = FastEstimatorDataLoader(pipeline)
            self.estimator.system.pipeline = pipeline

            runners[col] = FastEstimatorTaskRunner(
                estimator=self.estimator, data_loader=data_loader)
            runners[col].set_optimizer_treatment('CONTINUE_LOCAL')

            for trace in runners[col].estimator.system.traces:
                if isinstance(trace, BestModelSaver):
                    save_dir_path = f'{trace.save_dir}/{col}'
                    os.makedirs(save_dir_path, exist_ok=True)
                    save_dir[col] = save_dir_path

            data_path += 1

        # Create the collaborators
        collaborators = {collaborator: fx.create_collaborator(
            plan, collaborator, runners[collaborator], aggregator)
            for collaborator in plan.authorized_cols}

        model = None
        for round_num in range(self.rounds):
            for col in plan.authorized_cols:

                collaborator = collaborators[col]

                if round_num != 0:
                    # For FastEstimator Jupyter notebook, models must be
                    # saved in different directories (i.e. path must be
                    # reset here)

                    runners[col].estimator.system.load_state(
                        f'save/{col}_state')
                    runners[col].rebuild_model(round_num, model_states[col])

                # Reset the save directory if BestModelSaver is present
                # in traces
                for trace in runners[col].estimator.system.traces:
                    if isinstance(trace, BestModelSaver):
                        trace.save_dir = save_dir[col]

                collaborator.run_simulation()

                model_states[col] = runners[col].get_tensor_dict(
                    with_opt_vars=True)
                model = runners[col].model
                runners[col].estimator.system.save_state(f'save/{col}_state')

        # TODO This will return the model from the last collaborator,
        #  NOT the final aggregated model (though they should be similar).
        # There should be a method added to the aggregator that will load
        # the best model from disk and return it
        return model


def split_data(train, eva, test, rank, collaborator_count):
    """Split data into N parts, where N is the collaborator count."""
    if collaborator_count == 1:
        return train, eva, test

    fraction = [1.0 / float(collaborator_count)]
    fraction *= (collaborator_count - 1)

    # Expand the split list into individual parameters
    train_split = train.split(*fraction)
    eva_split = eva.split(*fraction)
    test_split = test.split(*fraction)

    train = [train]
    eva = [eva]
    test = [test]

    if type(train_split) is not list:
        train.append(train_split)
        eva.append(eva_split)
        test.append(test_split)
    else:
        # Combine all partitions into a single list
        train = [train] + train_split
        eva = [eva] + eva_split
        test = [test] + test_split

    # Extract the right shard
    train = train[rank - 1]
    eva = eva[rank - 1]
    test = test[rank - 1]

    return train, eva, test



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Python low-level API module."""
import os
import time
from collections import defaultdict
from copy import deepcopy
from logging import getLogger
from pathlib import Path
from typing import Dict
from typing import Tuple

from tensorboardX import SummaryWriter

from openfl.component.assigner.tasks import Task
from openfl.component.assigner.tasks import TrainTask
from openfl.component.assigner.tasks import ValidateTask
from openfl.federated import Plan
from openfl.interface.aggregation_functions import AggregationFunction
from openfl.interface.aggregation_functions import WeightedAverage
from openfl.interface.cli import setup_logging
from openfl.interface.cli_helper import WORKSPACE
from openfl.native import update_plan
from openfl.utilities import split_tensor_dict_for_holdouts
from openfl.utilities.workspace import dump_requirements_file


class ModelStatus:
    """Model statuses."""

    INITIAL = 'initial'
    BEST = 'best'
    LAST = 'last'
    RESTORED = 'restored'


class FLExperiment:
    """Central class for FL experiment orchestration."""

    def __init__(
            self,
            federation,
            experiment_name: str = None,
            serializer_plugin: str = 'openfl.plugins.interface_serializer.'
                                     'cloudpickle_serializer.CloudpickleSerializer',
            load_default_plan=True,
            nn=True
    ) -> None:
        """
        Initialize an experiment inside a federation.

        Experiment makes sense in a scope of some machine learning problem.
        Information about the data on collaborators is contained on the federation level.
        """
        self.federation = federation
        self.experiment_name = experiment_name or 'test-' + time.strftime('%Y%m%d-%H%M%S')
        self.summary_writer = None
        self.serializer_plugin = serializer_plugin
        self.load_default_plan = load_default_plan

        self.experiment_accepted = False

        self.is_validate_task_exist = False

        self.nn = nn

        self.logger = getLogger(__name__)
        setup_logging()

        self._initialize_plan()

    def _initialize_plan(self):
        """Setup plan from base plan interactive api."""
        # Create a folder to store plans
        os.makedirs('./plan', exist_ok=True)
        os.makedirs('./save', exist_ok=True)
        # Load the default plan
        if self.load_default_plan:
            base_plan_path = WORKSPACE / 'workspace/plan/plans/default/base_plan_interactive_api.yaml'
            plan = Plan.parse(base_plan_path, resolve=False)
        else:
            # TODO: make the path configuratble by the user
            plan = Plan.parse(Path("./plan/plan.yaml"), resolve=False)
        # Change plan name to default one
        plan.name = 'plan.yaml'

        self.plan = deepcopy(plan)

    def _assert_experiment_accepted(self):
        """Assure experiment is sent to director."""
        if not self.experiment_accepted:
            self.logger.error('The experiment has not been accepted by director')
            self.logger.error(
                'Report the experiment first: '
                'use the Experiment.start() method.')
            raise Exception

    def get_best_model(self):
        """Retrieve the model with the best score."""
        self._assert_experiment_accepted()
        tensor_dict = self.federation.dir_client.get_best_model(
            experiment_name=self.experiment_name)

        return self._rebuild_model(tensor_dict, upcoming_model_status=ModelStatus.BEST)

    def get_last_model(self):
        """Retrieve the aggregated model after the last round."""
        self._assert_experiment_accepted()
        tensor_dict = self.federation.dir_client.get_last_model(
            experiment_name=self.experiment_name)

        return self._rebuild_model(tensor_dict, upcoming_model_status=ModelStatus.LAST)

    def _rebuild_model(self, tensor_dict, upcoming_model_status=ModelStatus.BEST):
        """Use tensor dict to update model weights."""
        if len(tensor_dict) == 0:
            warning_msg = ('No tensors received from director\n'
                           'Possible reasons:\n'
                           '\t1. Aggregated model is not ready\n'
                           '\t2. Experiment data removed from director')

            if upcoming_model_status == ModelStatus.BEST and not self.is_validate_task_exist:
                warning_msg += '\n\t3. No validation tasks are provided'

            warning_msg += f'\nReturn {self.current_model_status} model'

            self.logger.warning(warning_msg)

        else:
            self.task_runner_stub.rebuild_model(tensor_dict, validation=True, device='cpu')
            self.current_model_status = upcoming_model_status

        return deepcopy(self.task_runner_stub.model)

    def stream_metrics(self, tensorboard_logs: bool = True) -> None:
        """Stream metrics."""
        self._assert_experiment_accepted()
        for metric_message_dict in self.federation.dir_client.stream_metrics(self.experiment_name):
            self.logger.metric(
                f'Round {metric_message_dict["round"]}, '
                f'collaborator {metric_message_dict["metric_origin"]} '
                f'{metric_message_dict["task_name"]} result '
                f'{metric_message_dict["metric_name"]}:\t{metric_message_dict["metric_value"]:f}')

            if tensorboard_logs:
                self.write_tensorboard_metric(metric_message_dict)

    def write_tensorboard_metric(self, metric: dict) -> None:
        """Write metric callback."""
        if not self.summary_writer:
            self.summary_writer = SummaryWriter(f'./logs/{self.experiment_name}', flush_secs=5)

        self.summary_writer.add_scalar(
            f'{metric["metric_origin"]}/{metric["task_name"]}/{metric["metric_name"]}',
            metric['metric_value'], metric['round'])

    def remove_experiment_data(self):
        """Remove experiment data."""
        self._assert_experiment_accepted()
        log_message = 'Removing experiment data '
        if self.federation.dir_client.remove_experiment_data(
                name=self.experiment_name
        ):
            log_message += 'succeed.'
            self.experiment_accepted = False
        else:
            log_message += 'failed.'

        self.logger.info(log_message)

    def prepare_workspace_distribution(self, model_provider, task_keeper, data_loader,
                                       task_assigner,
                                       pip_install_options: Tuple[str] = ()):
        """Prepare an archive from a user workspace."""
        # Save serialized python objects to disc
        self._serialize_interface_objects(model_provider, task_keeper, data_loader, task_assigner)
        # Save the prepared plan
        Plan.dump(Path(f'./plan/{self.plan.name}'), self.plan.config, freeze=False)

        # PACK the WORKSPACE!
        # Prepare requirements file to restore python env
        dump_requirements_file(keep_original_prefixes=True,
                               prefixes=pip_install_options)

        # Compress te workspace to restore it on collaborator
        self.arch_path = self._pack_the_workspace()

    def start(self, *, model_provider, task_keeper, data_loader,
              rounds_to_train: int,
              task_assigner=None,
              override_config: dict = None,
              delta_updates: bool = False,
              opt_treatment: str = 'RESET',
              device_assignment_policy: str = 'CPU_ONLY',
              pip_install_options: Tuple[str] = ()) -> None:
        """
        Prepare workspace distribution and send to Director.

        A successful call of this function will result in sending the experiment workspace
        to the Director service and experiment start.

        Parameters:
        model_provider - Model Interface instance.
        task_keeper - Task Interface instance.
        data_loader - Data Interface instance.
        rounds_to_train - required number of training rounds for the experiment.
        delta_updates - [bool] Tells if collaborators should send delta updates
            for the locally tuned models. If set to False, whole checkpoints will be sent.
        opt_treatment - Optimizer state treatment policy.
            Valid options: 'RESET' - reinitialize optimizer for every round,
            'CONTINUE_LOCAL' - keep local optimizer state,
            'CONTINUE_GLOBAL' - aggregate optimizer state.
        device_assignment_policy - device assignment policy.
            Valid options: 'CPU_ONLY' - device parameter passed to tasks
            will always be 'cpu',
            'CUDA_PREFERRED' - enable passing CUDA device identifiers to tasks
            by collaborators, works with cuda-device-monitor plugin equipped Envoys.
        pip_install_options - tuple of options for the remote `pip install` calls,
            example: ('-f some.website', '--no-index')
        """
        if not task_assigner:
            task_assigner = self.define_task_assigner(task_keeper, rounds_to_train)

        self._prepare_plan(model_provider, data_loader,
                           rounds_to_train,
                           delta_updates=delta_updates, opt_treatment=opt_treatment,
                           device_assignment_policy=device_assignment_policy,
                           override_config=override_config,
                           model_interface_file='model_obj.pkl',
                           tasks_interface_file='tasks_obj.pkl',
                           dataloader_interface_file='loader_obj.pkl')

        self.prepare_workspace_distribution(
            model_provider, task_keeper, data_loader,
            task_assigner,
            pip_install_options
        )

        self.logger.info('Starting experiment!')
        self.plan.resolve()
        initial_tensor_dict = self._get_initial_tensor_dict(model_provider, self.nn)
        try:
            response = self.federation.dir_client.set_new_experiment(
                name=self.experiment_name,
                col_names=self.plan.authorized_cols,
                arch_path=self.arch_path,
                initial_tensor_dict=initial_tensor_dict,
                nn=self.nn
            )
        finally:
            self.remove_workspace_archive()

        if response.accepted:
            self.logger.info('Experiment was accepted and launched.')
            self.experiment_accepted = True
        else:
            self.logger.info('Experiment was not accepted or failed.')

    def define_task_assigner(self, task_keeper, rounds_to_train):
        """Define task assigner by registered tasks."""
        tasks = task_keeper.get_registered_tasks()
        is_train_task_exist = False
        self.is_validate_task_exist = False
        for task in tasks.values():
            if task.task_type == 'train':
                is_train_task_exist = True
            if task.task_type == 'validate':
                self.is_validate_task_exist = True

        if not is_train_task_exist and rounds_to_train != 1:
            # Since we have only validation tasks, we do not have to train it multiple times
            raise Exception('Variable rounds_to_train must be equal 1, '
                            'because only validation tasks were given')
        if is_train_task_exist and self.is_validate_task_exist:
            def assigner(collaborators, round_number, **kwargs):
                tasks_by_collaborator = {}
                for collaborator in collaborators:
                    tasks_by_collaborator[collaborator] = [
                        tasks['train'],
                        tasks['locally_tuned_model_validate'],
                        tasks['aggregated_model_validate'],
                    ]
                return tasks_by_collaborator

            return assigner
        elif not is_train_task_exist and self.is_validate_task_exist:
            def assigner(collaborators, round_number, **kwargs):
                tasks_by_collaborator = {}
                for collaborator in collaborators:
                    tasks_by_collaborator[collaborator] = [
                        tasks['aggregated_model_validate'],
                    ]
                return tasks_by_collaborator

            return assigner
        elif is_train_task_exist and not self.is_validate_task_exist:
            raise Exception('You should define validate task!')
        else:
            raise Exception('You should define train and validate tasks!')

    def restore_experiment_state(self, model_provider):
        """Restore accepted experiment object."""
        self.task_runner_stub = self.plan.get_core_task_runner(model_provider=model_provider)
        self.current_model_status = ModelStatus.RESTORED
        self.experiment_accepted = True

    @staticmethod
    def _pack_the_workspace():
        """Packing the archive."""
        from shutil import copytree
        from shutil import ignore_patterns
        from shutil import make_archive
        from shutil import rmtree
        from os import getcwd
        from os import makedirs
        from os.path import basename

        archive_type = 'zip'
        archive_name = basename(getcwd())

        tmp_dir = 'temp_' + archive_name
        makedirs(tmp_dir, exist_ok=True)

        ignore = ignore_patterns(
            '__pycache__', 'data', 'cert', tmp_dir, '*.crt', '*.key',
            '*.csr', '*.srl', '*.pem', '*.pbuf', '*zip')

        copytree('./', tmp_dir + '/workspace', ignore=ignore)

        arch_path = make_archive(archive_name, archive_type, tmp_dir + '/workspace')

        rmtree(tmp_dir)

        return arch_path

    def remove_workspace_archive(self):
        """Remove the workspace archive."""
        os.remove(self.arch_path)
        del self.arch_path

    def _get_initial_tensor_dict(self, model_provider, nn):
        """Extract initial weights from the model."""
        self.task_runner_stub = self.plan.get_core_task_runner(model_provider=model_provider, nn=nn)
        self.current_model_status = ModelStatus.INITIAL
        if nn:
            tensor_dict, _ = split_tensor_dict_for_holdouts(
                self.logger,
                self.task_runner_stub.get_tensor_dict(False),
                **self.task_runner_stub.tensor_dict_split_fn_kwargs
            )
        else:
            tensor_dict = self.task_runner_stub.get_tensor_dict()

        return tensor_dict

    def _prepare_plan(self, model_provider, data_loader,
                      rounds_to_train,
                      delta_updates, opt_treatment,
                      device_assignment_policy,
                      override_config=None,
                      model_interface_file='model_obj.pkl', tasks_interface_file='tasks_obj.pkl',
                      dataloader_interface_file='loader_obj.pkl',
                      aggregation_function_interface_file='aggregation_function_obj.pkl',
                      task_assigner_file='task_assigner_obj.pkl'):
        """Fill plan.yaml file using user provided setting."""

        # Seems like we still need to fill authorized_cols list
        # So aggregator know when to start sending tasks
        # We also could change the aggregator logic so it will send tasks to aggregator
        # as soon as it connects. This change should be a part of a bigger PR
        # brining in fault tolerance changes

        shard_registry = self.federation.get_shard_registry()
        self.plan.authorized_cols = [
            name for name, info in shard_registry.items() if info['is_online']
        ]
        # Network part of the plan
        # We keep in mind that an aggregator FQND will be the same as the directors FQDN
        # We just choose a port randomly from plan hash
        director_fqdn = self.federation.director_node_fqdn.split(':')[0]  # We drop the port
        self.plan.config['network']['settings']['agg_addr'] = director_fqdn
        self.plan.config['network']['settings']['tls'] = self.federation.tls

        # Aggregator part of the plan
        self.plan.config['aggregator']['settings']['rounds_to_train'] = rounds_to_train

        # Collaborator part
        self.plan.config['collaborator']['settings']['delta_updates'] = delta_updates
        self.plan.config['collaborator']['settings']['opt_treatment'] = opt_treatment
        self.plan.config['collaborator']['settings'][
            'device_assignment_policy'] = device_assignment_policy

        # DataLoader part
        for setting, value in data_loader.kwargs.items():
            self.plan.config['data_loader']['settings'][setting] = value

        # TaskRunner framework plugin
        # ['required_plugin_components'] should be already in the default plan with all the fields
        # filled with the default values
        self.plan.config['task_runner']['required_plugin_components'] = {
            'framework_adapters': model_provider.framework_plugin
        }

        # API layer
        self.plan.config['api_layer'] = {
            'required_plugin_components': {
                'serializer_plugin': self.serializer_plugin
            },
            'settings': {
                'model_interface_file': model_interface_file,
                'tasks_interface_file': tasks_interface_file,
                'dataloader_interface_file': dataloader_interface_file,
                'aggregation_function_interface_file': aggregation_function_interface_file,
                'task_assigner_file': task_assigner_file
            }
        }

        if override_config:
            self.plan = update_plan(override_config, plan=self.plan, resolve=False)

    def _serialize_interface_objects(
            self,
            model_provider,
            task_keeper,
            data_loader,
            task_assigner
    ):
        """Save python objects to be restored on collaborators."""
        serializer = self.plan.build(
            self.plan.config['api_layer']['required_plugin_components']['serializer_plugin'], {})
        framework_adapter = Plan.build(model_provider.framework_plugin, {})
        # Model provider serialization may need preprocessing steps
        framework_adapter.serialization_setup()

        obj_dict = {
            'model_interface_file': model_provider,
            'tasks_interface_file': task_keeper,
            'dataloader_interface_file': data_loader,
            'aggregation_function_interface_file': task_keeper.aggregation_functions,
            'task_assigner_file': task_assigner
        }

        for filename, object_ in obj_dict.items():
            serializer.serialize(object_, self.plan.config['api_layer']['settings'][filename])


class TaskKeeper:
    """
    Task keeper class.

    Task should accept the following entities that exist on collaborator nodes:
    1. model - will be rebuilt with relevant weights for every task by `TaskRunner`
    2. data_loader - data loader equipped with `repository adapter` that provides local data
    3. device - a device to be used on collaborator machines
    4. optimizer (optional)

    Task returns a dictionary {metric name: metric value for this task}
    """

    def __init__(self) -> None:
        """Initialize task registry."""
        # Mapping 'task name' -> callable
        self.task_registry = {}
        # Mapping 'task name' -> arguments
        self.task_contract = {}
        # Mapping 'task name' -> arguments
        self.task_settings = defaultdict(dict)
        # Mapping 'task name' -> callable
        self.aggregation_functions = defaultdict(WeightedAverage)
        # Mapping 'task_alias' -> Task
        self._tasks: Dict[str, Task] = {}

    # TODO: maybe in case of AdaBoost.F the adaboost_coeff could be stored in the optimizer parameter
    def register_fl_task(self, model, data_loader, device, optimizer=None, round_num=None, adaboost_coeff=None,
                         name=None, nn=True):
        """
        Register FL tasks.

        The task contract should be set up by providing variable names:
        [model, data_loader, device] - necessarily
        and optimizer - optionally

        All tasks should accept contract entities to be run on collaborator node.
        Moreover we ask users return dict{'metric':value} in every task
        `
        TI = TaskInterface()

        task_settings = {
            'batch_size': 32,
            'some_arg': 228,
        }
        @TI.add_kwargs(**task_settings)
        @TI.register_fl_task(model='my_model', data_loader='train_loader',
                device='device', optimizer='my_Adam_opt')
        def foo_task(my_model, train_loader, my_Adam_opt, device, batch_size, some_arg=356)
            ...
            return {'metric_name': metric, 'metric_name_2': metric_2,}
        `
        """

        # The highest level wrapper for allowing arguments for the decorator
        def decorator_with_args(training_method):
            # We could pass hooks to the decorator
            # @functools.wraps(training_method)

            def wrapper_decorator(**task_keywords):
                metric_dict = training_method(**task_keywords)
                return metric_dict

            # Saving the task and the contract for later serialization
            function_name = training_method.__name__

            self.task_registry[function_name] = wrapper_decorator
            if nn:
                contract = {'model': model, 'data_loader': data_loader,
                            'device': device, 'optimizer': optimizer, 'round_num': round_num}
            else:
                contract = {'model': model, 'data_loader': data_loader,
                            'device': device, 'optimizer': optimizer, 'round_num': round_num,
                            'adaboost_coeff': adaboost_coeff, 'name': name}
            self.task_contract[function_name] = contract
            # define tasks
            if nn:
                if optimizer:
                    self._tasks['train'] = TrainTask(
                        name='train',
                        function_name=function_name,
                    )
                else:
                    self._tasks['locally_tuned_model_validate'] = ValidateTask(
                        name='locally_tuned_model_validate',
                        function_name=function_name,
                        apply_local=True,
                    )
                    self._tasks['aggregated_model_validate'] = ValidateTask(
                        name='aggregated_model_validate',
                        function_name=function_name,
                    )
                # We do not alter user environment
            else:
                if function_name == 'train_adaboost':
                    self._tasks['1_train'] = TrainTask(
                        name='1_train',
                        function_name=function_name,
                    )
                elif function_name == 'validate_weak_learners':
                    self._tasks['2_weak_learners_validate'] = ValidateTask(
                        name='2_weak_learners_validate',
                        function_name=function_name,
                        apply_local=True,
                    )
                elif function_name == 'adaboost_update':
                    self._tasks['3_adaboost_update'] = ValidateTask(
                        name='3_adaboost_update',
                        function_name=function_name,
                        apply_local=True,
                    )
                elif function_name == 'validate_adaboost':
                    self._tasks['4_adaboost_validate'] = ValidateTask(
                        name='4_adaboost_validate',
                        function_name=function_name,
                    )

            return training_method

        return decorator_with_args

    def add_kwargs(self, **task_kwargs):
        """
        Register tasks settings.

        Warning! We do not actually need to register additional kwargs,
        we ust serialize them.
        This one is a decorator because we need task name and
        to be consistent with the main registering method
        """

        # The highest level wrapper for allowing arguments for the decorator
        def decorator_with_args(training_method):
            # Saving the task's settings to be written in plan
            self.task_settings[training_method.__name__] = task_kwargs

            return training_method

        return decorator_with_args

    def set_aggregation_function(self, aggregation_function: AggregationFunction):
        """Set aggregation function for the task.

        To be serialized and sent to aggregator node.

        There is no support for aggregation functions
        containing logic from workspace-related libraries
        that are not present on director yet.

        Args:
            aggregation_function: Aggregation function.

        You might need to override default FedAvg aggregation with built-in aggregation types:
            - openfl.interface.aggregation_functions.GeometricMedian
            - openfl.interface.aggregation_functions.Median
        or define your own AggregationFunction subclass.
        See more details on `Overriding the aggregation function`_ documentation page.
        .. _Overriding the aggregation function:
            https://openfl.readthedocs.io/en/latest/overriding_agg_fn.html
        """

        def decorator_with_args(training_method):
            if not isinstance(aggregation_function, AggregationFunction):
                raise Exception('aggregation_function must implement '
                                'AggregationFunction interface.')
            self.aggregation_functions[training_method.__name__] = aggregation_function
            return training_method

        return decorator_with_args

    def get_registered_tasks(self) -> Dict[str, Task]:
        """Return registered tasks."""
        return self._tasks


# Backward compatibility
TaskInterface = TaskKeeper


class ModelInterface:
    """
    Registers model graph and optimizer.

    To be serialized and sent to collaborator nodes

    This is the place to determine correct framework adapter
        as they are needed to fill the model graph with trained tensors.

    There is no support for several models / optimizers yet.
    """

    def __init__(self, model, optimizer, framework_plugin) -> None:
        """
        Initialize model keeper.

        Tensors in provided graphs will be used for
        initialization of the global model.

        Arguments:
        model: Union[tuple, graph]
        optimizer: Union[tuple, optimizer]
        """
        self.model = model
        self.optimizer = optimizer
        self.framework_plugin = framework_plugin

    def provide_model(self):
        """Retrieve model."""
        return self.model

    def provide_optimizer(self):
        """Retrieve optimizer."""
        return self.optimizer


class DataInterface:
    """
    The class to define dataloaders.

    In the future users will have to adapt `unified data interface hook`
        in their dataloaders.
    For now, we can provide `data_path` variable on every collaborator node
        at initialization time for dataloader customization
    """

    def __init__(self, **kwargs):
        """Initialize DataLoader."""
        self.kwargs = kwargs

    @property
    def shard_descriptor(self):
        """Return shard descriptor."""
        return self._shard_descriptor

    @shard_descriptor.setter
    def shard_descriptor(self, shard_descriptor):
        """
        Describe per-collaborator procedures or sharding.

        This method will be called during a collaborator initialization.
        Local shard_descriptor  will be set by Envoy.
        """
        self._shard_descriptor = shard_descriptor
        raise NotImplementedError

    def get_train_loader(self, **kwargs):
        """Output of this method will be provided to tasks with optimizer in contract."""
        raise NotImplementedError

    def get_valid_loader(self, **kwargs):
        """Output of this method will be provided to tasks without optimizer in contract."""
        raise NotImplementedError

    def get_train_data_size(self):
        """Information for aggregation."""
        raise NotImplementedError

    def get_valid_data_size(self):
        """Information for aggregation."""
        raise NotImplementedError



from openfl.interface.interactive_api.experiment import DataInterface


class forestcoverDataset(DataInterface):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def shard_descriptor(self):
        return self._shard_descriptor

    @shard_descriptor.setter
    def shard_descriptor(self, shard_descriptor):
        """
        Describe per-collaborator procedures or sharding.

        This method will be called during a collaborator initialization.
        Local shard_descriptor  will be set by Envoy.
        """
        self._shard_descriptor = shard_descriptor

        self.train = shard_descriptor.get_dataset('train')
        self.test = shard_descriptor.get_dataset('val', complete=True)

    def __getitem__(self, index):
        return self.shard_descriptor[index]

    def __len__(self):
        return len(self.shard_descriptor)

    def get_train_loader(self, **kwargs):
        """
        Output of this method will be provided to tasks with optimizer in contract
        """
        return self.train.get_data()

    def get_valid_loader(self, **kwargs):
        """
        Output of this method will be provided to tasks without optimizer in contract
        """
        return self.test.get_data()

    def get_train_data_size(self):
        """
        Information for aggregation
        """
        return len(self.train)

    def get_valid_data_size(self):
        """
        Information for aggregation
        """
        return len(self.test)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Logs utilities."""

import logging

from rich.console import Console
from rich.logging import RichHandler
from tensorboardX import SummaryWriter

writer = None


def get_writer():
    """Create global writer object."""
    global writer
    if not writer:
        writer = SummaryWriter('./logs/tensorboard', flush_secs=5)


def write_metric(node_name, task_name, metric_name, metric, round_number):
    """Write metric callback."""
    get_writer()
    writer.add_scalar(f'{node_name}/{task_name}/{metric_name}', metric, round_number)


def setup_loggers(log_level=logging.INFO):
    """Configure loggers."""
    root = logging.getLogger()
    root.setLevel(log_level)
    console = Console(width=160)
    handler = RichHandler(console=console)
    formatter = logging.Formatter(
        '[%(asctime)s][%(name)s][%(levelname)s] - %(message)s'
    )
    handler.setFormatter(formatter)
    root.addHandler(handler)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl.transport.grpc package."""

from .aggregator_client import AggregatorGRPCClient
from .aggregator_server import AggregatorGRPCServer
from .director_server import DirectorGRPCServer

__all__ = [
    'AggregatorGRPCServer',
    'AggregatorGRPCClient',
    'DirectorGRPCServer',
]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Plan module."""

import sys
from logging import getLogger

from click import echo
from click import group
from click import option
from click import pass_context
from click import Path as ClickPath

from openfl.utilities.path_check import is_directory_traversal

logger = getLogger(__name__)


@group()
@pass_context
def plan(context):
    """Manage Federated Learning Plans."""
    context.obj['group'] = 'plan'


@plan.command()
@pass_context
@option('-p', '--plan_config', required=False,
        help='Federated learning plan [plan/plan.yaml]',
        default='plan/plan.yaml', type=ClickPath(exists=True))
@option('-c', '--cols_config', required=False,
        help='Authorized collaborator list [plan/cols.yaml]',
        default='plan/cols.yaml', type=ClickPath(exists=True))
@option('-d', '--data_config', required=False,
        help='The data set/shard configuration file [plan/data.yaml]',
        default='plan/data.yaml', type=ClickPath(exists=True))
@option('-a', '--aggregator_address', required=False,
        help='The FQDN of the federation agregator')
@option('-f', '--feature_shape', required=False,
        help='The input shape to the model')
def initialize(context, plan_config, cols_config, data_config,
               aggregator_address, feature_shape):
    """
    Initialize Data Science plan.

    Create a protocol buffer file of the initial model weights for
     the federation.
    """
    from pathlib import Path

    from openfl.federated import Plan
    from openfl.protocols import utils
    from openfl.utilities import split_tensor_dict_for_holdouts
    from openfl.utilities.utils import getfqdn_env

    for p in [plan_config, cols_config, data_config]:
        if is_directory_traversal(p):
            echo(f'{p} is out of the openfl workspace scope.')
            sys.exit(1)

    plan_config = Path(plan_config).absolute()
    cols_config = Path(cols_config).absolute()
    data_config = Path(data_config).absolute()

    plan = Plan.parse(plan_config_path=plan_config,
                      cols_config_path=cols_config,
                      data_config_path=data_config)

    init_state_path = plan.config['aggregator']['settings']['init_state_path']

    # TODO:  Is this part really needed?  Why would we need to collaborator
    #  name to know the input shape to the model?

    # if  feature_shape is None:
    #     if  cols_config is None:
    #         exit('You must specify either a feature
    #         shape or authorized collaborator
    #         list in order for the script to determine the input layer shape')
    print(plan.cols_data_paths)

    collaborator_cname = list(plan.cols_data_paths)[0]

    data_loader = plan.get_data_loader(collaborator_cname)
    task_runner = plan.get_task_runner(data_loader)
    tensor_pipe = plan.get_tensor_pipe()

    tensor_dict, holdout_params = split_tensor_dict_for_holdouts(
        logger,
        task_runner.get_tensor_dict(False),
        **task_runner.tensor_dict_split_fn_kwargs
    )

    logger.warn(f'Following parameters omitted from global initial model, '
                f'local initialization will determine'
                f' values: {list(holdout_params.keys())}')

    model_snap = utils.construct_model_proto(tensor_dict=tensor_dict,
                                             round_number=0,
                                             tensor_pipe=tensor_pipe)

    logger.info(f'Creating Initial Weights File    🠆 {init_state_path}')

    utils.dump_proto(model_proto=model_snap, fpath=init_state_path)

    plan_origin = Plan.parse(plan_config, resolve=False).config

    if (plan_origin['network']['settings']['agg_addr'] == 'auto'
            or aggregator_address):
        plan_origin['network']['settings']['agg_addr'] = aggregator_address or getfqdn_env()

        logger.warn(f'Patching Aggregator Addr in Plan'
                    f" 🠆 {plan_origin['network']['settings']['agg_addr']}")

        Plan.dump(plan_config, plan_origin)

    plan.config = plan_origin

    # Record that plan with this hash has been initialized
    if 'plans' not in context.obj:
        context.obj['plans'] = []
    context.obj['plans'].append(f'{plan_config.stem}_{plan.hash[:8]}')
    logger.info(f"{context.obj['plans']}")


# TODO: looks like Plan.method
def freeze_plan(plan_config):
    """Dump the plan to YAML file."""
    from pathlib import Path

    from openfl.federated import Plan

    plan = Plan()
    plan.config = Plan.parse(Path(plan_config), resolve=False).config

    init_state_path = plan.config['aggregator']['settings']['init_state_path']

    if not Path(init_state_path).exists():
        logger.info("Plan has not been initialized! Run 'fx plan"
                    " initialize' before proceeding")
        return

    Plan.dump(Path(plan_config), plan.config, freeze=True)


@plan.command(name='freeze')
@option('-p', '--plan_config', required=False,
        help='Federated learning plan [plan/plan.yaml]',
        default='plan/plan.yaml', type=ClickPath(exists=True))
def freeze(plan_config):
    """
    Finalize the Data Science plan.

    Create a new plan file that embeds its hash in the file name
    (plan.yaml -> plan_{hash}.yaml) and changes the permissions to read only
    """
    if is_directory_traversal(plan_config):
        echo('Plan config path is out of the openfl workspace scope.')
        sys.exit(1)
    freeze_plan(plan_config)


def switch_plan(name):
    """Switch the FL plan to this one."""
    from shutil import copyfile
    from os.path import isfile

    from yaml import dump
    from yaml import FullLoader
    from yaml import load

    plan_file = f'plan/plans/{name}/plan.yaml'
    if isfile(plan_file):

        echo(f'Switch plan to {name}')

        # Copy the new plan.yaml file to the top directory
        copyfile(plan_file, 'plan/plan.yaml')

        # Update the .workspace file to show the current workspace plan
        workspace_file = '.workspace'

        with open(workspace_file, 'r') as f:
            doc = load(f, Loader=FullLoader)

        if not doc:  # YAML is not correctly formatted
            doc = {}  # Create empty dictionary

        doc['current_plan_name'] = f'{name}'  # Switch with new plan name

        # Rewrite updated workspace file
        with open(workspace_file, 'w') as f:
            dump(doc, f)

    else:
        echo(f'Error: Plan {name} not found in plan/plans/{name}')


@plan.command(name='switch')
@option('-n', '--name', required=False,
        help='Name of the Federated learning plan',
        default='default', type=str)
def switch_(name):
    """Switch the current plan to this plan."""
    switch_plan(name)


@plan.command(name='save')
@option('-n', '--name', required=False,
        help='Name of the Federated learning plan',
        default='default', type=str)
def save_(name):
    """Save the current plan to this plan and switch."""
    from os import makedirs
    from shutil import copyfile

    echo(f'Saving plan to {name}')
    # TODO: How do we get the prefix path? What happens if this gets executed
    #  outside of the workspace top directory?

    makedirs(f'plan/plans/{name}', exist_ok=True)
    copyfile('plan/plan.yaml', f'plan/plans/{name}/plan.yaml')

    switch_plan(name)  # Swtich the context


@plan.command(name='remove')
@option('-n', '--name', required=False,
        help='Name of the Federated learning plan',
        default='default', type=str)
def remove_(name):
    """Remove this plan."""
    from shutil import rmtree

    if name != 'default':
        echo(f'Removing plan {name}')
        # TODO: How do we get the prefix path? What happens if
        #  this gets executed outside of the workspace top directory?

        rmtree(f'plan/plans/{name}')

        switch_plan('default')  # Swtich the context back to the default

    else:
        echo("ERROR: Can't remove default plan")


@plan.command(name='print')
def print_():
    """Print the current plan."""
    from openfl.interface.cli_helper import get_workspace_parameter

    current_plan_name = get_workspace_parameter('current_plan_name')
    echo(f'The current plan is: {current_plan_name}')



import numpy as np


class AdaBoostF:
    def __init__(self, base_estimator, n_classes):
        self.estimators_ = [base_estimator]
        self.n_estimators_ = 1
        self.estimator_weights_ = [1]
        self.n_classes = n_classes

    def get_estimators(self):
        return self.estimators_

    def add(self, weak_learner, coeff):
        self.estimators_.append(weak_learner)
        self.estimator_weights_ = np.append(self.estimator_weights_, coeff)

    def get(self, index):
        return self.estimators_[index]

    def replace(self, weak_learner, coeff):
        self.estimators_ = [weak_learner]
        self.estimator_weights_ = np.array([coeff])

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        y_pred = np.zeros((np.shape(X)[0], self.n_classes))
        for i, clf in enumerate(self.estimators_):
            pred = clf.predict(X)
            for j, c in enumerate(pred):
                y_pred[j, int(c)] += self.estimator_weights_[i]
        return np.argmax(y_pred, axis=1)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow as tf

from openfl.federated import KerasTaskRunner
from .define_model import build_model
from .define_model import dice_coef
from .define_model import dice_loss
from .define_model import soft_dice_coef


class TensorFlow3dUNet(KerasTaskRunner):
    """Initialize.

    Args:
        **kwargs: Additional parameters to pass to the function

    """

    def __init__(self, initial_filters=16,
                 depth=5,
                 batch_norm=True,
                 use_upsampling=False,
                 **kwargs):
        """Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function

        """
        super().__init__(**kwargs)

        self.model = self.create_model(
            input_shape=self.feature_shape,
            n_cl_out=self.data_loader.num_classes,
            initial_filters=initial_filters,
            use_upsampling=use_upsampling,
            depth=depth,
            batch_norm=batch_norm,
            **kwargs
        )
        self.initialize_tensorkeys_for_functions()

        self.model.summary(print_fn=self.logger.info, line_length=120)

    def create_model(self,
                     input_shape,
                     n_cl_out=1,
                     use_upsampling=False,
                     dropout=0.2,
                     print_summary=True,
                     seed=816,
                     depth=5,
                     dropout_at=(2, 3),
                     initial_filters=16,
                     batch_norm=True,
                     **kwargs):
        """Create the TensorFlow 3D U-Net CNN model.

        Args:
            input_shape (list): input shape of the data
            n_cl_out (int): Number of output classes in label (Default=1)
            **kwargs: Additional parameters to pass to the function

        """
        #
        # Define Model
        #
        model = build_model(input_shape,
                            n_cl_out=n_cl_out,
                            use_upsampling=use_upsampling,
                            dropout=dropout,
                            print_summary=print_summary,
                            seed=seed,
                            depth=depth,
                            dropout_at=dropout_at,
                            initial_filters=initial_filters,
                            batch_norm=batch_norm)

        self.optimizer = tf.keras.optimizers.Adam()

        model.compile(
            loss=dice_loss,
            optimizer=self.optimizer,
            metrics=[dice_coef, soft_dice_coef],
        )

        self.tvars = model.layers
        print(f'layer names: {[var.name for var in self.tvars]}')

        self.opt_vars = self.optimizer.variables()
        print(f'optimizer vars: {self.opt_vars}')

        # Two opt_vars for one tvar: gradient and square sum for RMSprop.
        self.fl_vars = self.tvars + self.opt_vars

        return model


if __name__ == '__main__':

    from tf_brats_dataloader import DatasetGenerator
    import os

    import argparse

    parser = argparse.ArgumentParser(
        description='Train 3D U-Net model', add_help=True,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--data_path',
                        default='~/data/MICCAI_BraTS2020_TrainingData/',
                        # Or wherever you unzipped the BraTS datset,
                        help='Root directory for BraTS 2020 dataset')
    parser.add_argument('--epochs',
                        type=int,
                        default=5,
                        help='Number of epochs')
    parser.add_argument('--crop_dim',
                        type=int,
                        default=64,
                        help='Crop all dimensions to this (height, width, depth)')
    parser.add_argument('--batch_size',
                        type=int,
                        default=4,
                        help='Training batch size')
    parser.add_argument('--train_test_split',
                        type=float,
                        default=0.80,
                        help='Train/test split (0-1)')
    parser.add_argument('--validate_test_split',
                        type=float,
                        default=0.50,
                        help='Validation/test split (0-1)')
    parser.add_argument('--number_input_channels',
                        type=int,
                        default=1,
                        help='Number of input channels')
    parser.add_argument('--num_classes',
                        type=int,
                        default=1,
                        help='Number of output classes/channels')
    parser.add_argument('--random_seed',
                        default=816,
                        help='Random seed for determinism')
    parser.add_argument('--print_model',
                        action='store_true',
                        default=True,
                        help='Print the summary of the model layers')
    parser.add_argument('--filters',
                        type=int,
                        default=16,
                        help='Number of filters in the first convolutional layer')
    parser.add_argument('--use_upsampling',
                        action='store_true',
                        default=False,
                        help='Use upsampling instead of transposed convolution')
    parser.add_argument('--use_batchnorm',
                        action='store_true',
                        default=True,
                        help='Use batch normalization')
    parser.add_argument('--saved_model_name',
                        default='saved_model_3DUnet',
                        help='Save model to this path')

    args = parser.parse_args()

    print(args)

    brats_data = DatasetGenerator(args.crop_dim,
                                  data_path=os.path.abspath(os.path.expanduser(args.data_path)),
                                  batch_size=args.batch_size,
                                  train_test_split=args.train_test_split,
                                  validate_test_split=args.validate_test_split,
                                  number_input_channels=args.number_input_channels,
                                  num_classes=args.num_classes,
                                  random_seed=args.random_seed
                                  )

    model = build_model([args.crop_dim, args.crop_dim, args.crop_dim, args.number_input_channels],
                        use_upsampling=args.use_upsampling,
                        n_cl_out=args.num_classes,
                        dropout=0.2,
                        print_summary=args.print_model,
                        seed=args.random_seed,
                        depth=5,
                        dropout_at=[2, 3],
                        initial_filters=args.filters,
                        batch_norm=args.use_batchnorm
                        )

    model.compile(loss=dice_loss,
                  optimizer=tf.keras.optimizers.Adam(learning_rate=0.01),
                  metrics=[dice_coef, soft_dice_coef]
                  )

    checkpoint = tf.keras.callbacks.ModelCheckpoint(args.saved_model_name,
                                                    verbose=1,
                                                    save_best_only=True)

    # TensorBoard
    import datetime
    logs_dir = os.path.join('tensorboard_logs',
                            datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))
    tb_logs = tf.keras.callbacks.TensorBoard(log_dir=logs_dir)

    callbacks = [checkpoint, tb_logs]

    history = model.fit(brats_data.ds_train,
                        validation_data=brats_data.ds_val,
                        epochs=args.epochs,
                        callbacks=callbacks)



from openfl.component.aggregation_functions.interface import AggregationFunction


class Identity(AggregationFunction):
    """Identity function for faking aggregation"""

    def call(self, local_tensors, *_):
        return local_tensors


# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Serializer plugin interface."""


class Serializer:
    """Serializer API plugin."""

    def __init__(self) -> None:
        """Initialize serializer."""
        pass

    @staticmethod
    def serialize(object_, filename):
        """Serialize an object and save to disk."""
        raise NotImplementedError

    @staticmethod
    def restore_object(filename):
        """Load and deserialize an object."""
        raise NotImplementedError



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Database Utilities."""

from .dataframe import _search,_store,_retrieve,ROUND_PLACEHOLDER,ROUND_PLACEHOLDER

__all__ = [
    '_search',
    '_store',
    '_retrieve',
    'ROUND_PLACEHOLDER'
]



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Mnist Shard Descriptor."""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from torchvision import datasets

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)


class MnistShardDataset(ShardDataset):
    """Mnist Shard dataset class."""

    def __init__(self, x, y, data_type, rank: int = 1, worldsize: int = 1) -> None:
        """Initialize Mnist shard Dataset."""
        self.data_type = data_type
        self.rank = rank
        self.worldsize = worldsize

        self.x = x[self.rank - 1::self.worldsize]
        self.y = y[self.rank - 1::self.worldsize]

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """Return an item by the index."""
        return self.x[index], self.y[index]

    def __len__(self) -> int:
        """Return the len of the dataset."""
        return len(self.x)


class MnistShardDescriptor(ShardDescriptor):
    """Mnist Shard descriptor class."""

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            **kwargs
    ) -> None:
        """Initialize MnistShardDescriptor."""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        (x_train, y_train), (x_val, y_val) = self.download_data()
        self.data_by_type = {
            'train': (x_train, y_train),
            'val': (x_val, y_val)
        }

    def get_shard_dataset_types(self) -> List[Dict[str, Any]]:
        """Get available shard dataset types."""
        return list(self.data_by_type)

    def get_dataset(self, dataset_type: str = 'train') -> MnistShardDataset:
        """Return a shard dataset by type."""
        if dataset_type not in self.data_by_type:
            raise Exception(f'Wrong dataset type: {dataset_type}')
        return MnistShardDataset(
            *self.data_by_type[dataset_type],
            data_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        return ['28', '28']

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        return ['1']

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Mnist dataset, shard number {self.rank}'
                f' out of {self.worldsize}')

    def download_data(self) -> Tuple[Tuple[Any, Any], Tuple[Any, Any]]:
        """Download prepared dataset."""
        train_data, val_data = (
            datasets.MNIST('data', train=train, download=True)
            for train in (True, False)
        )
        x_train, y_train = train_data.train_data, train_data.train_labels
        x_val, y_val = val_data.test_data, val_data.test_labels

        print('Mnist data was loaded!')
        return (x_train, y_train), (x_val, y_val)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""
Shard Descriptor template.

It is recommended to perform tensor manipulations using numpy.
"""

from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor


class LocalShardDescriptor(ShardDescriptor):
    """Shard descriptor subclass."""

    def __init__(self, data_path: str, sample_shape: tuple, target_shape: tuple) -> None:
        """
        Initialize local Shard Descriptor.

        Parameters are arbitrary, set up the ShardDescriptor-related part
        of the envoy_config.yaml as you need.
        """
        super().__init__()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Pytorch Framework Adapter plugin."""
from copy import deepcopy

import numpy as np
import torch as pt

from .framework_adapter_interface import FrameworkAdapterPluginInterface


class FrameworkAdapterPlugin(FrameworkAdapterPluginInterface):
    """Framework adapter plugin class."""

    def __init__(self) -> None:
        """Initialize framework adapter."""
        pass

    @staticmethod
    def get_tensor_dict(model, optimizer=None):
        """
        Extract tensor dict from a model and an optimizer.

        Returns:
        dict {weight name: numpy ndarray}
        """
        state = to_cpu_numpy(model.state_dict())

        if optimizer is not None:
            opt_state = _get_optimizer_state(optimizer)
            state = {**state, **opt_state}

        return state

    @staticmethod
    def set_tensor_dict(model, tensor_dict, optimizer=None, device='cpu'):
        """
        Set tensor dict from a model and an optimizer.

        Given a dict {weight name: numpy ndarray} sets weights to
        the model and optimizer objects inplace.
        """
        new_state = {}
        # Grabbing keys from model's state_dict helps to confirm we have
        # everything
        for k in model.state_dict():
            new_state[k] = pt.from_numpy(tensor_dict.pop(k)).to(device)

        # set model state
        model.load_state_dict(new_state)

        if optimizer is not None:
            # see if there is state to restore first
            if tensor_dict.pop('__opt_state_needed') == 'true':
                _set_optimizer_state(optimizer, device, tensor_dict)

            # sanity check that we did not record any state that was not used
            assert len(tensor_dict) == 0


def _set_optimizer_state(optimizer, device, derived_opt_state_dict):
    """Set the optimizer state.

    Args:
        optimizer:
        device:
        derived_opt_state_dict:

    """
    temp_state_dict = expand_derived_opt_state_dict(
        derived_opt_state_dict, device)

    # Setting other items from the param_groups
    # getting them from the local optimizer
    # (expand_derived_opt_state_dict sets only 'params')
    for i, group in enumerate(optimizer.param_groups):
        for k, v in group.items():
            if k not in temp_state_dict['param_groups'][i]:
                temp_state_dict['param_groups'][i][k] = v

    optimizer.load_state_dict(temp_state_dict)


def _get_optimizer_state(optimizer):
    """Return the optimizer state.

    Args:
        optimizer
    """
    opt_state_dict = deepcopy(optimizer.state_dict())

    # Optimizer state might not have some parts representing frozen parameters
    # So we do not synchronize them
    param_keys_with_state = set(opt_state_dict['state'].keys())
    for group in opt_state_dict['param_groups']:
        local_param_set = set(group['params'])
        params_to_sync = local_param_set & param_keys_with_state
        group['params'] = sorted(params_to_sync)

    derived_opt_state_dict = _derive_opt_state_dict(opt_state_dict)

    return derived_opt_state_dict


def _derive_opt_state_dict(opt_state_dict):
    """Separate optimizer tensors from the tensor dictionary.

    Flattens the optimizer state dict so as to have key, value pairs with
    values as numpy arrays.
    The keys have sufficient info to restore opt_state_dict using
    expand_derived_opt_state_dict.

    Args:
        opt_state_dict: The optimizer state dictionary

    """
    derived_opt_state_dict = {}

    # Determine if state is needed for this optimizer.
    if len(opt_state_dict['state']) == 0:
        derived_opt_state_dict['__opt_state_needed'] = 'false'
        return derived_opt_state_dict

    derived_opt_state_dict['__opt_state_needed'] = 'true'

    # Using one example state key, we collect keys for the corresponding
    # dictionary value.
    example_state_key = opt_state_dict['param_groups'][0]['params'][0]
    example_state_subkeys = set(
        opt_state_dict['state'][example_state_key].keys()
    )

    # We assume that the state collected for all params in all param groups is
    # the same.
    # We also assume that whether or not the associated values to these state
    # subkeys is a tensor depends only on the subkey.
    # Using assert statements to break the routine if these assumptions are
    # incorrect.
    for state_key in opt_state_dict['state'].keys():
        assert example_state_subkeys == set(opt_state_dict['state'][state_key].keys())
        for state_subkey in example_state_subkeys:
            assert (isinstance(
                opt_state_dict['state'][example_state_key][state_subkey],
                pt.Tensor)
                == isinstance(
                    opt_state_dict['state'][state_key][state_subkey],
                    pt.Tensor))

    state_subkeys = list(opt_state_dict['state'][example_state_key].keys())

    # Tags will record whether the value associated to the subkey is a
    # tensor or not.
    state_subkey_tags = []
    for state_subkey in state_subkeys:
        if isinstance(
                opt_state_dict['state'][example_state_key][state_subkey],
                pt.Tensor
        ):
            state_subkey_tags.append('istensor')
        else:
            state_subkey_tags.append('')
    state_subkeys_and_tags = list(zip(state_subkeys, state_subkey_tags))

    # Forming the flattened dict, using a concatenation of group index,
    # subindex, tag, and subkey inserted into the flattened dict key -
    # needed for reconstruction.
    nb_params_per_group = []
    for group_idx, group in enumerate(opt_state_dict['param_groups']):
        for idx, param_id in enumerate(group['params']):
            for subkey, tag in state_subkeys_and_tags:
                if tag == 'istensor':
                    new_v = opt_state_dict['state'][param_id][
                        subkey].cpu().numpy()
                else:
                    new_v = np.array(
                        [opt_state_dict['state'][param_id][subkey]]
                    )
                derived_opt_state_dict[f'__opt_state_{group_idx}_{idx}_{tag}_{subkey}'] = new_v
        nb_params_per_group.append(idx + 1)
    # group lengths are also helpful for reconstructing
    # original opt_state_dict structure
    derived_opt_state_dict['__opt_group_lengths'] = np.array(
        nb_params_per_group
    )

    return derived_opt_state_dict


def expand_derived_opt_state_dict(derived_opt_state_dict, device):
    """Expand the optimizer state dictionary.

    Takes a derived opt_state_dict and creates an opt_state_dict suitable as
    input for load_state_dict for restoring optimizer state.

    Reconstructing state_subkeys_and_tags using the example key
    prefix, "__opt_state_0_0_", certain to be present.

    Args:
        derived_opt_state_dict: Optimizer state dictionary

    Returns:
        dict: Optimizer state dictionary
    """
    state_subkeys_and_tags = []
    for key in derived_opt_state_dict:
        if key.startswith('__opt_state_0_0_'):
            stripped_key = key[16:]
            if stripped_key.startswith('istensor_'):
                this_tag = 'istensor'
                subkey = stripped_key[9:]
            else:
                this_tag = ''
                subkey = stripped_key[1:]
            state_subkeys_and_tags.append((subkey, this_tag))

    opt_state_dict = {'param_groups': [], 'state': {}}
    nb_params_per_group = list(
        derived_opt_state_dict.pop('__opt_group_lengths').astype(np.int)
    )

    # Construct the expanded dict.
    for group_idx, nb_params in enumerate(nb_params_per_group):
        these_group_ids = [
            f'{group_idx}_{idx}' for idx in range(nb_params)
        ]
        opt_state_dict['param_groups'].append({'params': these_group_ids})
        for this_id in these_group_ids:
            opt_state_dict['state'][this_id] = {}
            for subkey, tag in state_subkeys_and_tags:
                flat_key = f'__opt_state_{this_id}_{tag}_{subkey}'
                if tag == 'istensor':
                    new_v = pt.from_numpy(derived_opt_state_dict.pop(flat_key))
                else:
                    # Here (for currrently supported optimizers) the subkey
                    # should be 'step' and the length of array should be one.
                    assert subkey == 'step'
                    assert len(derived_opt_state_dict[flat_key]) == 1
                    new_v = int(derived_opt_state_dict.pop(flat_key))
                opt_state_dict['state'][this_id][subkey] = new_v

    # sanity check that we did not miss any optimizer state
    assert len(derived_opt_state_dict) == 0, str(derived_opt_state_dict)

    return opt_state_dict


def to_cpu_numpy(state):
    """Send data to CPU as Numpy array.

    Args:
        state

    """
    # deep copy so as to decouple from active model
    state = deepcopy(state)

    for k, v in state.items():
        # When restoring, we currently assume all values are tensors.
        if not pt.is_tensor(v):
            raise ValueError('We do not currently support non-tensors '
                             'coming from model.state_dict()')
        # get as a numpy array, making sure is on cpu
        state[k] = v.cpu().numpy()
    return state



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""


import os

from openfl.federated import TensorFlowDataLoader
from .dataloader import DatasetGenerator


class TensorFlowBratsDataLoader(TensorFlowDataLoader):
    """TensorFlow Data Loader for the BraTS dataset."""

    def __init__(self, data_path, batch_size=4,
                 crop_dim=64, percent_train=0.8,
                 pre_split_shuffle=True,
                 number_input_channels=1,
                 num_classes=1,
                 **kwargs):
        """Initialize.

        Args:
            data_path: The file path for the BraTS dataset
            batch_size (int): The batch size to use
            crop_dim (int): Crop the original image to this size on each dimension
            percent_train (float): The percentage of the data to use for training (Default=0.8)
            pre_split_shuffle (bool): True= shuffle the dataset before
            performing the train/validate split (Default=True)
            **kwargs: Additional arguments, passed to super init

        Returns:
            Data loader with BraTS data
        """
        super().__init__(batch_size, **kwargs)

        self.data_path = os.path.abspath(os.path.expanduser(data_path))
        self.batch_size = batch_size
        self.crop_dim = [crop_dim, crop_dim, crop_dim, number_input_channels]
        self.num_input_channels = number_input_channels
        self.num_classes = num_classes

        self.train_test_split = percent_train

        self.brats_data = DatasetGenerator(crop_dim,
                                           data_path=data_path,
                                           number_input_channels=number_input_channels,
                                           batch_size=batch_size,
                                           train_test_split=percent_train,
                                           validate_test_split=0.5,
                                           num_classes=num_classes,
                                           random_seed=816)

    def get_feature_shape(self):
        """
        Get the shape of an example feature array.

        Returns:
            tuple: shape of an example feature array
        """
        return tuple(self.brats_data.get_input_shape())

    def get_train_loader(self, batch_size=None, num_batches=None):
        """
        Get training data loader.

        Returns
        -------
        loader object
        """
        return self.brats_data.ds_train

    def get_valid_loader(self, batch_size=None):
        """
        Get validation data loader.

        Returns:
            loader object
        """
        return self.brats_data.ds_val

    def get_train_data_size(self):
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        return self.brats_data.num_train

    def get_valid_data_size(self):
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        return self.brats_data.num_val



# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Custom model DeviceArray - JAX Numpy adapter."""
import jax
import jax.numpy as jnp
import numpy as np
import optax

from flax import traverse_util
from .framework_adapter_interface import FrameworkAdapterPluginInterface

class FrameworkAdapterPlugin(FrameworkAdapterPluginInterface):
    """Framework adapter plugin class."""

    @staticmethod
    def get_tensor_dict(model, optimizer=None):
        """
        Extract tensor dict from a model.params and model.opt_state (optimizer).

        Returns:
        dict {weight name: numpy ndarray}

        """

        # Convert PyTree Structure DeviceArray to Numpy
        model_params = jax.tree_util.tree_map(np.array, model.params)
        params_dict = _get_weights_dict(model_params, 'param')
        
        # If optimizer is initialized
        # Optax Optimizer agnostic state processing (TraceState, AdamScaleState, any...)
        if not isinstance(model.opt_state[0], optax.EmptyState):
            opt_state = jax.tree_util.tree_map(np.array, model.opt_state)[0]
            opt_vars = filter(_get_opt_vars, dir(opt_state))
            for var in opt_vars:
                opt_dict = getattr(opt_state, var) # Returns a dict
                # Flattens a deeply nested dictionary
                opt_dict = _get_weights_dict(opt_dict, f'opt_{var}')
                params_dict.update(opt_dict)

        return params_dict

    @staticmethod
    def set_tensor_dict(model, tensor_dict, optimizer=None, device='cpu'):
        """
        Set the `model.params and model.opt_state` with a flattened tensor dictionary.
        Choice of JAX platform (device) cpu/gpu/gpu is initialized at start.
        Args:
            tensor_dict: flattened {weight name: numpy ndarray} tensor dictionary

        Returns:
            None
        """
        
        tensor_dict = jax.tree_util.tree_map(jnp.array, tensor_dict)

        _set_weights_dict(model, tensor_dict, 'param')

        if not isinstance(model.opt_state[0], optax.EmptyState):
            _set_weights_dict(model, tensor_dict, 'opt')

def _get_opt_vars(x):
    return False if x.startswith('_') or x in ['index', 'count'] else True

def _set_weights_dict(obj, weights_dict, prefix=''):
    """Set the object weights with a dictionary.

    The obj can be a model or an optimizer.
    
    Args:
        obj (Model or Optimizer): The target object that we want to set
        the weights.
        weights_dict (dict): The weight dictionary.

    Returns:
        None
    """
    
    if prefix == 'opt':
        model_state_dict = obj.opt_state[0]
        # opt_vars -> ['mu', 'nu'] for Adam or ['trace'] for SGD or ['ANY'] for any
        opt_vars = filter(_get_opt_vars, dir(model_state_dict))
        for var in opt_vars:
            opt_state_dict = getattr(model_state_dict, var)
            _update_weights(opt_state_dict, weights_dict, prefix, var)
    else:
        _update_weights(obj.params, weights_dict, prefix)

def _update_weights(state_dict, tensor_dict, prefix, suffix=None):
    # Re-assignment of the state variable(s) is restricted.
    # Instead update the nested layers weights iteratively.
    dict_prefix = f'{prefix}_{suffix}' if suffix is not None else f'{prefix}'
    for layer_name, param_obj in state_dict.items():
        for param_name, value in param_obj.items():
            key = '*'.join([dict_prefix, layer_name, param_name])
            if key in tensor_dict:
                state_dict[layer_name][param_name] = tensor_dict[key]

def _get_weights_dict(obj, prefix):
    """
    Get the dictionary of weights.

    Parameters
    ----------
    obj : Model or Optimizer
        The target object that we want to get the weights.

    Returns
    -------
    dict
        The weight dictionary.
    """
    weights_dict = {prefix: obj}
    # Flatten the dictionary with a given separator for
    # easy lookup and assignment in `set_tensor_dict` method.
    flat_params = traverse_util.flatten_dict(weights_dict, sep='*')
    return flat_params


# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Mnist Shard Descriptor."""

import logging
from typing import List

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor
from sklearn.datasets import load_svmlight_file
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

logger = logging.getLogger(__name__)


class krvskpShardDataset(ShardDataset):
    """Mnist Shard dataset class."""

    def __init__(self, x, y, data_type, rank=1, worldsize=1, complete=False):
        """Initialize TinyImageNetDataset."""
        self.data_type = data_type
        self.rank = rank
        self.worldsize = worldsize
        self.x = x if complete else x[self.rank - 1::self.worldsize]
        self.y = y if complete else y[self.rank - 1::self.worldsize]

    def get_data(self):
        return self.x, self.y

    def __getitem__(self, index: int):
        """Return an item by the index."""
        return self.x[index], self.y[index]

    def __len__(self):
        """Return the len of the dataset."""
        return len(self.x)


class krvskpShardDescriptor(ShardDescriptor):
    """Mnist Shard descriptor class."""

    def __init__(
            self,
            rank_worldsize: str = '1, 1',
            **kwargs
    ):
        """Initialize MnistShardDescriptor."""
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        x_train, x_test, y_train, y_test = self.download_data()

        self.data_by_type = {
            'train': (x_train, y_train),
            'val': (x_test, y_test)
        }

    def get_shard_dataset_types(self) -> List[str]:
        """Get available shard dataset types."""
        return list(self.data_by_type)

    def get_dataset(self, dataset_type='train', complete=False):
        """Return a shard dataset by type."""
        if dataset_type not in self.data_by_type:
            raise Exception(f'Wrong dataset type: {dataset_type}')
        return krvskpShardDataset(
            *self.data_by_type[dataset_type],
            data_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize,
            complete=complete
        )

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return ['36']

    @property
    def target_shape(self):
        """Return the target shape info."""
        return ['1']

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Adult dataset, shard number {self.rank}'
                f' out of {self.worldsize}')

    def download_data(self):
        X, y = load_svmlight_file("../krvskp_dataset/kr-vs-kp.svmlight")
        X = X.toarray()
        y = y.astype("int")
        y = LabelEncoder().fit_transform(y)

        return train_test_split(X, y, test_size=0.2)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from openfl.protocols import director_pb2 as openfl_dot_protocols_dot_director__pb2


class DirectorStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.UpdateShardInfo = channel.unary_unary(
                '/openfl.director.Director/UpdateShardInfo',
                request_serializer=openfl_dot_protocols_dot_director__pb2.UpdateShardInfoRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.UpdateShardInfoResponse.FromString,
                )
        self.WaitExperiment = channel.stream_stream(
                '/openfl.director.Director/WaitExperiment',
                request_serializer=openfl_dot_protocols_dot_director__pb2.WaitExperimentRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.WaitExperimentResponse.FromString,
                )
        self.GetExperimentData = channel.unary_stream(
                '/openfl.director.Director/GetExperimentData',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetExperimentDataRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.ExperimentData.FromString,
                )
        self.UpdateEnvoyStatus = channel.unary_unary(
                '/openfl.director.Director/UpdateEnvoyStatus',
                request_serializer=openfl_dot_protocols_dot_director__pb2.UpdateEnvoyStatusRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.UpdateEnvoyStatusResponse.FromString,
                )
        self.SetExperimentFailed = channel.unary_unary(
                '/openfl.director.Director/SetExperimentFailed',
                request_serializer=openfl_dot_protocols_dot_director__pb2.SetExperimentFailedRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.SetExperimentFailedResponse.FromString,
                )
        self.GetExperimentDescription = channel.unary_unary(
                '/openfl.director.Director/GetExperimentDescription',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetExperimentDescriptionRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.GetExperimentDescriptionResponse.FromString,
                )
        self.GetExperimentsList = channel.unary_unary(
                '/openfl.director.Director/GetExperimentsList',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetExperimentsListRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.GetExperimentsListResponse.FromString,
                )
        self.SetNewExperiment = channel.stream_unary(
                '/openfl.director.Director/SetNewExperiment',
                request_serializer=openfl_dot_protocols_dot_director__pb2.ExperimentInfo.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.SetNewExperimentResponse.FromString,
                )
        self.GetDatasetInfo = channel.unary_unary(
                '/openfl.director.Director/GetDatasetInfo',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetDatasetInfoRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.GetDatasetInfoResponse.FromString,
                )
        self.GetTrainedModel = channel.unary_unary(
                '/openfl.director.Director/GetTrainedModel',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetTrainedModelRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.TrainedModelResponse.FromString,
                )
        self.GetMetricStream = channel.unary_stream(
                '/openfl.director.Director/GetMetricStream',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetMetricStreamRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.GetMetricStreamResponse.FromString,
                )
        self.RemoveExperimentData = channel.unary_unary(
                '/openfl.director.Director/RemoveExperimentData',
                request_serializer=openfl_dot_protocols_dot_director__pb2.RemoveExperimentRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.RemoveExperimentResponse.FromString,
                )
        self.GetEnvoys = channel.unary_unary(
                '/openfl.director.Director/GetEnvoys',
                request_serializer=openfl_dot_protocols_dot_director__pb2.GetEnvoysRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_director__pb2.GetEnvoysResponse.FromString,
                )


class DirectorServicer(object):
    """Missing associated documentation comment in .proto file."""

    def UpdateShardInfo(self, request, context):
        """1. Envoy RPCs
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WaitExperiment(self, request_iterator, context):
        """Shard owner could also provide some public data for tests
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetExperimentData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateEnvoyStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetExperimentFailed(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetExperimentDescription(self, request, context):
        """2. Frontend RPCs
        2.1 Extension RPCs
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetExperimentsList(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetNewExperiment(self, request_iterator, context):
        """2.2 API RPCs
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDatasetInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrainedModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMetricStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveExperimentData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEnvoys(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DirectorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'UpdateShardInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateShardInfo,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.UpdateShardInfoRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.UpdateShardInfoResponse.SerializeToString,
            ),
            'WaitExperiment': grpc.stream_stream_rpc_method_handler(
                    servicer.WaitExperiment,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.WaitExperimentRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.WaitExperimentResponse.SerializeToString,
            ),
            'GetExperimentData': grpc.unary_stream_rpc_method_handler(
                    servicer.GetExperimentData,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetExperimentDataRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.ExperimentData.SerializeToString,
            ),
            'UpdateEnvoyStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateEnvoyStatus,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.UpdateEnvoyStatusRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.UpdateEnvoyStatusResponse.SerializeToString,
            ),
            'SetExperimentFailed': grpc.unary_unary_rpc_method_handler(
                    servicer.SetExperimentFailed,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.SetExperimentFailedRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.SetExperimentFailedResponse.SerializeToString,
            ),
            'GetExperimentDescription': grpc.unary_unary_rpc_method_handler(
                    servicer.GetExperimentDescription,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetExperimentDescriptionRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.GetExperimentDescriptionResponse.SerializeToString,
            ),
            'GetExperimentsList': grpc.unary_unary_rpc_method_handler(
                    servicer.GetExperimentsList,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetExperimentsListRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.GetExperimentsListResponse.SerializeToString,
            ),
            'SetNewExperiment': grpc.stream_unary_rpc_method_handler(
                    servicer.SetNewExperiment,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.ExperimentInfo.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.SetNewExperimentResponse.SerializeToString,
            ),
            'GetDatasetInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDatasetInfo,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetDatasetInfoRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.GetDatasetInfoResponse.SerializeToString,
            ),
            'GetTrainedModel': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrainedModel,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetTrainedModelRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.TrainedModelResponse.SerializeToString,
            ),
            'GetMetricStream': grpc.unary_stream_rpc_method_handler(
                    servicer.GetMetricStream,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetMetricStreamRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.GetMetricStreamResponse.SerializeToString,
            ),
            'RemoveExperimentData': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveExperimentData,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.RemoveExperimentRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.RemoveExperimentResponse.SerializeToString,
            ),
            'GetEnvoys': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEnvoys,
                    request_deserializer=openfl_dot_protocols_dot_director__pb2.GetEnvoysRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_director__pb2.GetEnvoysResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'openfl.director.Director', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Director(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def UpdateShardInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/UpdateShardInfo',
            openfl_dot_protocols_dot_director__pb2.UpdateShardInfoRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.UpdateShardInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def WaitExperiment(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/openfl.director.Director/WaitExperiment',
            openfl_dot_protocols_dot_director__pb2.WaitExperimentRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.WaitExperimentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetExperimentData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/openfl.director.Director/GetExperimentData',
            openfl_dot_protocols_dot_director__pb2.GetExperimentDataRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.ExperimentData.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateEnvoyStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/UpdateEnvoyStatus',
            openfl_dot_protocols_dot_director__pb2.UpdateEnvoyStatusRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.UpdateEnvoyStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetExperimentFailed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/SetExperimentFailed',
            openfl_dot_protocols_dot_director__pb2.SetExperimentFailedRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.SetExperimentFailedResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetExperimentDescription(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/GetExperimentDescription',
            openfl_dot_protocols_dot_director__pb2.GetExperimentDescriptionRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.GetExperimentDescriptionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetExperimentsList(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/GetExperimentsList',
            openfl_dot_protocols_dot_director__pb2.GetExperimentsListRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.GetExperimentsListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetNewExperiment(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/openfl.director.Director/SetNewExperiment',
            openfl_dot_protocols_dot_director__pb2.ExperimentInfo.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.SetNewExperimentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDatasetInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/GetDatasetInfo',
            openfl_dot_protocols_dot_director__pb2.GetDatasetInfoRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.GetDatasetInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrainedModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/GetTrainedModel',
            openfl_dot_protocols_dot_director__pb2.GetTrainedModelRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.TrainedModelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMetricStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/openfl.director.Director/GetMetricStream',
            openfl_dot_protocols_dot_director__pb2.GetMetricStreamRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.GetMetricStreamResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveExperimentData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/RemoveExperimentData',
            openfl_dot_protocols_dot_director__pb2.RemoveExperimentRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.RemoveExperimentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetEnvoys(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/openfl.director.Director/GetEnvoys',
            openfl_dot_protocols_dot_director__pb2.GetEnvoysRequest.SerializeToString,
            openfl_dot_protocols_dot_director__pb2.GetEnvoysResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Collaborator module."""

import sys
from logging import getLogger

from click import echo
from click import group
from click import option
from click import pass_context
from click import Path as ClickPath
from click import style

from openfl.utilities.path_check import is_directory_traversal

logger = getLogger(__name__)


@group()
@pass_context
def collaborator(context):
    """Manage Federated Learning Collaborators."""
    context.obj['group'] = 'service'


@collaborator.command(name='start')
@option('-p', '--plan', required=False,
        help='Federated learning plan [plan/plan.yaml]',
        default='plan/plan.yaml',
        type=ClickPath(exists=True))
@option('-d', '--data_config', required=False,
        help='The data set/shard configuration file [plan/data.yaml]',
        default='plan/data.yaml', type=ClickPath(exists=True))
@option('-n', '--collaborator_name', required=True,
        help='The certified common name of the collaborator')
@option('-s', '--secure', required=False,
        help='Enable Intel SGX Enclave', is_flag=True, default=False)
def start_(plan, collaborator_name, data_config, secure):
    """Start a collaborator service."""
    from pathlib import Path

    from openfl.federated import Plan

    if plan and is_directory_traversal(plan):
        echo('Federated learning plan path is out of the openfl workspace scope.')
        sys.exit(1)
    if data_config and is_directory_traversal(data_config):
        echo('The data set/shard configuration file path is out of the openfl workspace scope.')
        sys.exit(1)

    plan = Plan.parse(plan_config_path=Path(plan).absolute(),
                      data_config_path=Path(data_config).absolute())

    # TODO: Need to restructure data loader config file loader

    echo(f'Data = {plan.cols_data_paths}')
    logger.info('🧿 Starting a Collaborator Service.')

    plan.get_collaborator(collaborator_name).run()


def register_data_path(collaborator_name, data_path=None, silent=False):
    """Register dataset path in the plan/data.yaml file.

    Args:
        collaborator_name (str): The collaborator whose data path to be defined
        data_path (str)        : Data path (optional)
        silent (bool)          : Silent operation (don't prompt)
    """
    from click import prompt
    from os.path import isfile

    if data_path and is_directory_traversal(data_path):
        echo('Data path is out of the openfl workspace scope.')
        sys.exit(1)

    # Ask for the data directory
    default_data_path = f'data/{collaborator_name}'
    if not silent and data_path is None:
        dir_path = prompt('\nWhere is the data (or what is the rank)'
                          ' for collaborator '
                          + style(f'{collaborator_name}', fg='green')
                          + ' ? ', default=default_data_path)
    elif data_path is not None:
        dir_path = data_path
    else:
        # TODO: Need to figure out the default for this.
        dir_path = default_data_path

    # Read the data.yaml file
    d = {}
    data_yaml = 'plan/data.yaml'
    separator = ','
    if isfile(data_yaml):
        with open(data_yaml, 'r') as f:
            for line in f:
                if separator in line:
                    key, val = line.split(separator, maxsplit=1)
                    d[key] = val.strip()

    d[collaborator_name] = dir_path

    # Write the data.yaml
    with open(data_yaml, 'w') as f:
        for key, val in d.items():
            f.write(f'{key}{separator}{val}\n')


@collaborator.command(name='generate-cert-request')
@option('-n', '--collaborator_name', required=True,
        help='The certified common name of the collaborator')
@option('-d', '--data_path',
        help='The data path to be associated with the collaborator')
@option('-s', '--silent', help='Do not prompt', is_flag=True)
@option('-x', '--skip-package',
        help='Do not package the certificate signing request for export',
        is_flag=True)
def generate_cert_request_(collaborator_name,
                           data_path, silent, skip_package):
    """Generate certificate request for the collaborator."""
    if data_path and is_directory_traversal(data_path):
        echo('Data path is out of the openfl workspace scope.')
        sys.exit(1)
    generate_cert_request(collaborator_name, data_path, silent, skip_package)


def generate_cert_request(collaborator_name, data_path, silent, skip_package):
    """
    Create collaborator certificate key pair.

    Then create a package with the CSR to send for signing.
    """
    from openfl.cryptography.participant import generate_csr
    from openfl.cryptography.io import write_crt
    from openfl.cryptography.io import write_key
    from openfl.interface.cli_helper import CERT_DIR

    common_name = f'{collaborator_name}'.lower()
    subject_alternative_name = f'DNS:{common_name}'
    file_name = f'col_{common_name}'

    echo(f'Creating COLLABORATOR certificate key pair with following settings: '
         f'CN={style(common_name, fg="red")},'
         f' SAN={style(subject_alternative_name, fg="red")}')

    client_private_key, client_csr = generate_csr(common_name, server=False)

    (CERT_DIR / 'client').mkdir(parents=True, exist_ok=True)

    echo('  Moving COLLABORATOR certificate to: ' + style(
        f'{CERT_DIR}/{file_name}', fg='green'))

    # Write collaborator csr and key to disk
    write_crt(client_csr, CERT_DIR / 'client' / f'{file_name}.csr')
    write_key(client_private_key, CERT_DIR / 'client' / f'{file_name}.key')

    if not skip_package:
        from shutil import copytree
        from shutil import ignore_patterns
        from shutil import make_archive
        from tempfile import mkdtemp
        from os.path import basename
        from os.path import join
        from os import remove
        from glob import glob

        archive_type = 'zip'
        archive_name = f'col_{common_name}_to_agg_cert_request'
        archive_file_name = archive_name + '.' + archive_type

        # Collaborator certificate signing request
        tmp_dir = join(mkdtemp(), 'openfl', archive_name)

        ignore = ignore_patterns('__pycache__', '*.key', '*.srl', '*.pem')
        # Copy the current directory into the temporary directory
        copytree(f'{CERT_DIR}/client', tmp_dir, ignore=ignore)

        for f in glob(f'{tmp_dir}/*'):
            if common_name not in basename(f):
                remove(f)

        # Create Zip archive of directory
        make_archive(archive_name, archive_type, tmp_dir)

        echo(f'Archive {archive_file_name} with certificate signing'
             f' request created')
        echo('This file should be sent to the certificate authority'
             ' (typically hosted by the aggregator) for signing')

    # TODO: There should be some association with the plan made here as well
    register_data_path(common_name, data_path=data_path, silent=silent)


def find_certificate_name(file_name):
    """Parse the collaborator name."""
    col_name = str(file_name).split('/')[-1].split('.')[0][4:]
    return col_name


def register_collaborator(file_name):
    """Register the collaborator name in the cols.yaml list.

    Args:
        file_name (str): The name of the collaborator in this federation

    """
    from os.path import isfile
    from yaml import dump
    from yaml import FullLoader
    from yaml import load
    from pathlib import Path

    col_name = find_certificate_name(file_name)

    cols_file = Path('plan/cols.yaml').absolute()

    if not isfile(cols_file):
        cols_file.touch()
    with open(cols_file, 'r') as f:
        doc = load(f, Loader=FullLoader)

    if not doc:  # YAML is not correctly formatted
        doc = {}  # Create empty dictionary

    # List doesn't exist
    if 'collaborators' not in doc.keys() or not doc['collaborators']:
        doc['collaborators'] = []  # Create empty list

    if col_name in doc['collaborators']:

        echo('\nCollaborator '
             + style(f'{col_name}', fg='green')
             + ' is already in the '
             + style(f'{cols_file}', fg='green'))

    else:

        doc['collaborators'].append(col_name)
        with open(cols_file, 'w') as f:
            dump(doc, f)

        echo('\nRegistering '
             + style(f'{col_name}', fg='green')
             + ' in '
             + style(f'{cols_file}', fg='green'))


@collaborator.command(name='certify')
@option('-n', '--collaborator_name',
        help='The certified common name of the collaborator. This is only'
             ' needed for single node expiriments')
@option('-s', '--silent', help='Do not prompt', is_flag=True)
@option('-r', '--request-pkg', type=ClickPath(exists=True),
        help='The archive containing the certificate signing'
             ' request (*.zip) for a collaborator')
@option('-i', '--import', 'import_', type=ClickPath(exists=True),
        help='Import the archive containing the collaborator\'s'
             ' certificate (signed by the CA)')
def certify_(collaborator_name, silent, request_pkg, import_):
    """Certify the collaborator."""
    certify(collaborator_name, silent, request_pkg, import_)


def certify(collaborator_name, silent, request_pkg=None, import_=False):
    """Sign/certify collaborator certificate key pair."""
    from click import confirm
    from pathlib import Path
    from shutil import copy
    from shutil import make_archive
    from shutil import unpack_archive
    from glob import glob
    from os.path import basename
    from os.path import join
    from os.path import splitext
    from os import remove
    from tempfile import mkdtemp
    from openfl.cryptography.ca import sign_certificate
    from openfl.cryptography.io import read_crt
    from openfl.cryptography.io import read_csr
    from openfl.cryptography.io import read_key
    from openfl.cryptography.io import write_crt
    from openfl.interface.cli_helper import CERT_DIR

    common_name = f'{collaborator_name}'.lower()

    if not import_:
        if request_pkg:
            Path(f'{CERT_DIR}/client').mkdir(parents=True, exist_ok=True)
            unpack_archive(request_pkg, extract_dir=f'{CERT_DIR}/client')
            csr = glob(f'{CERT_DIR}/client/*.csr')[0]
        else:
            if collaborator_name is None:
                echo('collaborator_name can only be omitted if signing\n'
                     'a zipped request package.\n'
                     '\n'
                     'Example: fx collaborator certify --request-pkg '
                     'col_one_to_agg_cert_request.zip')
                return
            csr = glob(f'{CERT_DIR}/client/col_{common_name}.csr')[0]
            copy(csr, CERT_DIR)
        cert_name = splitext(csr)[0]
        file_name = basename(cert_name)
        signing_key_path = 'ca/signing-ca/private/signing-ca.key'
        signing_crt_path = 'ca/signing-ca.crt'

        # Load CSR
        if not Path(f'{cert_name}.csr').exists():
            echo(style('Collaborator certificate signing request not found.', fg='red')
                 + ' Please run `fx collaborator generate-cert-request`'
                   ' to generate the certificate request.')

        csr, csr_hash = read_csr(f'{cert_name}.csr')

        # Load private signing key
        if not Path(CERT_DIR / signing_key_path).exists():
            echo(style('Signing key not found.', fg='red')
                 + ' Please run `fx workspace certify`'
                   ' to initialize the local certificate authority.')

        signing_key = read_key(CERT_DIR / signing_key_path)

        # Load signing cert
        if not Path(CERT_DIR / signing_crt_path).exists():
            echo(style('Signing certificate not found.', fg='red')
                 + ' Please run `fx workspace certify`'
                   ' to initialize the local certificate authority.')

        signing_crt = read_crt(CERT_DIR / signing_crt_path)

        echo('The CSR Hash for file '
             + style(f'{file_name}.csr', fg='green')
             + ' = '
             + style(f'{csr_hash}', fg='red'))

        if silent:

            echo(' Signing COLLABORATOR certificate')
            signed_col_cert = sign_certificate(csr, signing_key, signing_crt.subject)
            write_crt(signed_col_cert, f'{cert_name}.crt')
            register_collaborator(CERT_DIR / 'client' / f'{file_name}.crt')

        else:

            if confirm('Do you want to sign this certificate?'):

                echo(' Signing COLLABORATOR certificate')
                signed_col_cert = sign_certificate(csr, signing_key, signing_crt.subject)
                write_crt(signed_col_cert, f'{cert_name}.crt')
                register_collaborator(CERT_DIR / 'client' / f'{file_name}.crt')

            else:
                echo(style('Not signing certificate.', fg='red')
                     + ' Please check with this collaborator to get the'
                       ' correct certificate for this federation.')
                return

        if len(common_name) == 0:
            # If the collaborator name is provided, the collaborator and
            # certificate does not need to be exported
            return

        # Remove unneeded CSR
        remove(f'{cert_name}.csr')

        archive_type = 'zip'
        archive_name = f'agg_to_{file_name}_signed_cert'

        # Collaborator certificate signing request
        tmp_dir = join(mkdtemp(), 'openfl', archive_name)

        Path(f'{tmp_dir}/client').mkdir(parents=True, exist_ok=True)
        # Copy the signed cert to the temporary directory
        copy(f'{CERT_DIR}/client/{file_name}.crt', f'{tmp_dir}/client/')
        # Copy the CA certificate chain to the temporary directory
        copy(f'{CERT_DIR}/cert_chain.crt', tmp_dir)

        # Create Zip archive of directory
        make_archive(archive_name, archive_type, tmp_dir)

    else:
        # Copy the signed certificate and cert chain into PKI_DIR
        previous_crts = glob(f'{CERT_DIR}/client/*.crt')
        unpack_archive(import_, extract_dir=CERT_DIR)
        updated_crts = glob(f'{CERT_DIR}/client/*.crt')
        cert_difference = list(set(updated_crts) - set(previous_crts))
        if len(cert_difference) != 0:
            crt = basename(cert_difference[0])
            echo(f'Certificate {crt} installed to PKI directory')
        else:
            echo('Certificate updated in the PKI directory')
